/***************************************************************************
                          PlayerVK.C  -  description
                             -------------------
    begin                : Thu Apr 17 2003
    copyright            : (C) 2003 by Vadim Kyrylov
    email                : vkyrylov@wct
 ***************************************************************************/

#include "Player.h" 
#include "Geometry.h"

//  A modified version by Vadim
//  This method tests if the opponent goal is undefended or more
//  sophisticated decision making is necessary by calling another method.
//  It returns true if the agent should score the opponent goal.
//  Besides, if the returned value is false, the method finds the recommended action type.
//    \param 'pos' is the returned best scoring point
//    \param 'action' is the recommended action type to execute.
bool Player::shouldIScore( VecPosition *pos, ActionT *action )
{
  bool shoot = false;
  VecPosition mypos;

  VecPosition posBall     = WM->getGlobalPosition( OBJECT_BALL );
  VecPosition posGoalie   = WM->predictPlayerPos( 3, WM->getOppGoalieType() );
  double goalDist         = posBall.getDistanceTo( WM->getPosOpponentGoal() );
  double delta            = 0.5 + goalDist/30.0;
  VecPosition leftCorner  = VecPosition( PITCH_LENGTH/2.0, -7.0+delta );    // a little right from the exact location
  VecPosition rightCorner = VecPosition( PITCH_LENGTH/2.0,  7.0-delta );    // a little left  from the exact location

  AngDeg      scoringDirection;
  AngDeg      goalieRelDir     = ( posGoalie - posBall ).getDirection();
  AngDeg      goalCenterRelDir = ( WM->getPosOpponentGoal() - posBall ).getDirection();
  AngDeg      leftCornerRelDir  = ( leftCorner - posBall ).getDirection();
  AngDeg      rightCornerRelDir = ( rightCorner - posBall ).getDirection();

  bool couldBeFreeGoal = false;
  ActionT myaction = ACT_ILLEGAL;                   // set it by default


  if ( WM->getOppGoalieType() != OBJECT_ILLEGAL )
  {
    if ( WM->getConfidence( WM->getOppGoalieType() ) >= 0.9 )
    {
      // I am fairly well sure about the opponent goalie whereabouts;
      // hence, more sophisticated decision making is necessary.
      //cout << "- checking if to shoot in spite of the goalie presense " << endl;
      shoot = shouldIScoreInSpiteOfGoalie( &mypos, &myaction );
    }
    else
    {
      // my info about the opponent goalie is outdated
      //cout << "- going to shoot, because my info about the opponent goalie is outdated " << endl;
      couldBeFreeGoal = true;
    }
  }
  else
  {
    // I have no idea of the opponent goalie
    //cout << "- going to shoot, because I have no idea of the opponent goalie " << endl;
    couldBeFreeGoal = true;
  }

  if ( couldBeFreeGoal )
  {
    AngDeg goalAngularSize = VecPosition::normalizeAngle( fabs( rightCornerRelDir - leftCornerRelDir ) );

    if (   goalDist < 15.0                               ||
         ( goalDist < 24.0  &&  goalAngularSize > 15.0 )    )
    {
      // find out which goal pole is closer and determine the scoring direction
      double distToLgoalCorner = posBall.getDistanceTo( leftCorner  );
      double distToRgoalCorner = posBall.getDistanceTo( rightCorner );

      if ( distToRgoalCorner > distToLgoalCorner )
        scoringDirection = VecPosition::normalizeAngle( ( leftCorner  - posBall ).getDirection() );
      else
        scoringDirection = VecPosition::normalizeAngle( ( rightCorner - posBall ).getDirection() );

      // calculate the scoring point
      mypos = VecPosition( WM->getPosOpponentGoal().getX(),
                           posBall.getY() + ( WM->getPosOpponentGoal().getX() - posBall.getX() ) *
                                                                         tanDeg( scoringDirection ) );
      if ( goalAngularSize > 15.0 )
      {
          shoot = true;
          myaction = ACT_KICK_TO;
          //cout << "#" << WM->getPlayerNumber() << " shooting at wide open free goal, goalAngularSize=" << goalAngularSize << ", goalDist=" << goalDist << endl;
      }
      else if ( goalDist < 15.0  )
      {
        // try to shoot in a narrow angle, or switch to an alternative action
        shoot = shouldIShootNarrowly( goalAngularSize, leftCorner, rightCorner, &mypos, &myaction );
      }
    }
  }

  if ( shoot )
  {
    // move the scoring point significantly behind the goal, to ensure a stronger kick
    mypos = mypos + ( mypos - posBall);
    *pos = mypos;
    //cout << "scoring point: " << mypos << endl;
  }

  *action = myaction;

  return shoot;

}


//  This version is used for backward compatibility only
bool Player::shouldIScore( VecPosition *pos )
{
  ActionT myaction;
  VecPosition mypos;
  bool should = shouldIScore( &mypos, &myaction );
  *pos = mypos;
  return should;
}



//  This method returns true if it makes sense to shoot the goal whose
//  angular size is rather narrow.
//  If it does not make sense to score, this method returns false. In this case it can
//  also calculate the recommended alternatoive action type.
//    \param 'anglWidth'   is the scoring angle width
//    \param 'leftCorner'  is the goal left corner position
//    \param 'rightCorner' is the goal right corner position
//    \param 'pos' is the returned best scoring point
//    \param 'action' is the recommended action type to execute
bool Player::shouldIShootNarrowly( AngDeg anglWidth, VecPosition leftCorner,
                                   VecPosition rightCorner, VecPosition *pos, ActionT *action )
{
  bool should = false;
  ActionT myaction = ACT_ILLEGAL;
  VecPosition mypos;

  VecPosition posBall    = WM->getGlobalPosition( OBJECT_BALL );
  double goalCornerDist;
  double distToLgoalCorner = posBall.getDistanceTo( leftCorner );
  double distToRgoalCorner = posBall.getDistanceTo( rightCorner );

  // use distance to the nearest goalCorner
  if ( distToRgoalCorner > distToLgoalCorner )
    goalCornerDist = distToLgoalCorner;
   else
    goalCornerDist = distToRgoalCorner;

  if ( goalCornerDist < 7.0 )                 // I am close to the goal corner
  {
    if ( anglWidth > 7.0 )                    // the scoring angle is rather wide
    {
      mypos =  WM->getPosOpponentGoal();

      if ( WM->isItSafeToPass( mypos, 10.0 ) )
      {
        // there are no opponents in the direction of the scoring point
        should = true;
        myaction = ACT_KICK_TO;
        //cout << "#" << WM->getPlayerNumber() << " shooting, in a narrow angle, goalAngularSize=" << anglWidth << endl;
      }
    }
    else
    {
      // check if it is safe to dribble several cycles from the goalline
      int iOpponent = 100;
      ObjectT fastestOpponent = WM->getFastestInSetTo( OBJECT_SET_OPPONENTS, OBJECT_BALL, &iOpponent );
      //cout << "checking if I can dribble, goalAngularSize=" << anglWidth << ", iOpponent=" << iOpponent << endl;
      if ( iOpponent > 5 )
      {
        myaction = ACT_DRIBBLE;
        //cout << "...dribbling instead of shooting, goalAngularSize=" << anglWidth << endl;
      }
    }
  }

  *action = myaction;
  *pos = mypos;
  return should;

}


//  This method tests if the opponent goal is undefended or more
//  sophisticated decision making is necessary by calling another method.
//  If it does not make sense to score, this method returns false. In this case it can
//  also calculate the recommended alternatoive action type.
//    \param 'pos' is the returned best scoring point
//    \param 'action' is the recommended action type to execute indeed when this function returns true
bool Player::shouldIScoreInSpiteOfGoalie( VecPosition *pos, ActionT *action )
{
  bool        shoot    = false;
  ActionT     myaction = ACT_ILLEGAL;
  VecPosition mypos;

  VecPosition posBall     = WM->getGlobalPosition( OBJECT_BALL );
  VecPosition posGoalie   = WM->predictPlayerPos( 3, WM->getOppGoalieType() );
  double goalDist         = posBall.getDistanceTo( WM->getPosOpponentGoal() );
  double delta            = 0.5 + goalDist/30.0;
  VecPosition leftCorner  = VecPosition ( PITCH_LENGTH/2.0, -7.0+delta );   // 0.9m right from the exacxt location
  VecPosition rightCorner = VecPosition( PITCH_LENGTH/2.0,  7.0-delta );   // 0.9m left  from the exacxt location

  AngDeg      scoringMaxAngle;
  AngDeg      goalieRelDir      = ( posGoalie                - posBall ).getDirection();
  AngDeg      goalCenterRelDir  = ( WM->getPosOpponentGoal() - posBall ).getDirection();
  AngDeg      leftCornerRelDir  = ( leftCorner               - posBall ).getDirection();
  AngDeg      rightCornerRelDir = ( rightCorner              - posBall ).getDirection();


  // select the scoring point in the farest corner from the goalie
  if ( fabs( goalieRelDir - leftCornerRelDir ) > fabs( rightCornerRelDir - goalieRelDir ) )
  {
    // goalie direction is closer to the RIGHT goal corner direction
    scoringMaxAngle = VecPosition::normalizeAngle( fabs( goalieRelDir - leftCornerRelDir ) );
    mypos = leftCorner;
  }
  else
  {
    // goalie direction is closer to the LEFT goal corner direction
    scoringMaxAngle = VecPosition::normalizeAngle( fabs( rightCornerRelDir - goalieRelDir ) );
    mypos = rightCorner;
  }

  double goalieDist = posBall.getDistanceTo( posGoalie );

  if ( goalDist < 24.0 )
  {
    if ( ( goalieRelDir < leftCornerRelDir  ||
           goalieRelDir > rightCornerRelDir    ) &&              // goalie is standing clear of the scoring angle, and
         scoringMaxAngle > 30.0                   )              // the scoring angle is rather wide
    {
      //cout << "#" << WM->getPlayerNumber() <<" ** shoot at free goal;  goalDist=" << goalDist << endl;
      shoot = true;
    }

    else if ( scoringMaxAngle > 45.0 )                           // the scoring angle is wide enough
    {
      //cout << "#" << WM->getPlayerNumber() <<" ** shoot at wide angle;  scoringMaxAngle=" << scoringMaxAngle << ";  goalDist=" << goalDist << endl;
      shoot = true;
      *pos = WM->getPointBetweenTwoClosestOpponentsInCone( posBall, leftCorner, rightCorner, 20.0 );
    }

    else if ( goalieDist > 4.0                                           &&
              goalieDist < 7.0                                           &&
//              fabs( posGoalie.getX() - WM->getPosOpponentGoal().getX() ) < 8.0  &&    // goalie is close to the goal center
//              fabs( posGoalie.getY() - WM->getPosOpponentGoal().getY() ) < 8.0  &&
              shouldITrickGoalie()                                          )
    {
      myaction = ACT_TRICK;
      //cout << "#" << WM->getPlayerNumber() << " ** tricking the goalie, not shooting;  goalieDist=" << goalieDist << endl;
    }

    else if ( ( goalDist < 15.0  || goalieDist < 15.0 ) &&
              scoringMaxAngle > 30.0                       )
    {
      //cout << "#" << WM->getPlayerNumber() <<" ** shoot at smaller angle;  scoringMaxAngle=" << scoringMaxAngle << ";  goalDist=" << goalDist << endl;
      shoot = true;
    }

    else if ( goalDist < 10.0  ||  goalieDist < 7.0 )
    {
      if ( WM->getAgentGlobalPosition().getY() < 0.0 )
        mypos = leftCorner;
      else
        mypos = rightCorner;

      shoot = true;
      //cout << "#" << WM->getPlayerNumber() <<" ** shoot at small distance to goal/goalie" << endl;
    }

    else if ( goalieDist > 5.0 &&
              shouldIShootNarrowly( scoringMaxAngle, leftCorner, rightCorner, &mypos, &myaction ) )
    {
      // I am close to a goal pole and the scoring angle is rather narrow
      // (possibly I will have to dribble with the ball a little)
      shoot = true;
    }
  }

  *pos = mypos;
  *action = myaction;

  return shoot;
}




//!  This method returns true if the agent should intercept the ball, either as
//   the fastest teammate or a backup.
//   Redeveloped from Serguei's version by Vadim
bool Player::shouldIIntercept( InterceptionT *iType )
{
  bool should = false;

  VecPosition posBall = WM->getBallPos();

  // the fastest player and his number of cycles
  int iFastest;
  ObjectT fastestPlayer;

  int iOpponent = 100;

  ObjectT fastestOpponent = WM->getFastestInSetTo( OBJECT_SET_OPPONENTS, OBJECT_BALL, &iOpponent );
  // fastest teammate and number of cycles
  int iTeammate = 100;
  ObjectT fastestTeammate = WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTeammate );
  // second fastest teammate and number of cycles
  int iTeammate1 = 100;
  ObjectT secondFastestTeammate = WM->getSecondFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTeammate1 );

  PlayerT fastestTeammateType = formations->getPlayerType( SoccerTypes::getIndex( fastestTeammate ) );
  PlayerT secondFastestTeammateType = formations->getPlayerType( SoccerTypes::getIndex( secondFastestTeammate ) );

  if ( iOpponent < iTeammate )
  {
     fastestPlayer = fastestOpponent;
     iFastest = iOpponent;
  }
  else
  {
     fastestPlayer = fastestTeammate;
     iFastest = iTeammate;
  }

  if ( not WM->isGoalKickUs() )
  {
    // If I am the fastest player, and threfore I must intercept
    if( fastestPlayer == WM->getAgentObjectType() )
    {
/*
      if ( WM->getPlayerNumber() <= 3 )
        cout << "...#" << WM->getPlayerNumber()
             << " I am the fastest PLAYER "
             << " iFast=" << iFastest
             << " iTmm=" << iTeammate1 << " #" << secondFastestTeammate - 60
             << " iOpp=" << iOpponent
             << " cycle=" << WM->getCurrentCycle()
             << endl;
*/
      should = true;
    }

    else if ( fastestTeammate == WM->getAgentObjectType() )
    {
      // I am the fastest teammate, and I must untercept except...
      should = true;
/*
      if ( WM->getPlayerNumber() <= 3 )
        cout << "...#" << WM->getPlayerNumber()
             << " I am the fastest TEAMMATE  "
             << " iFast=" << iFastest
             << " iTmm=" << iTeammate1 << " #" << secondFastestTeammate - 60
             << " iOpp=" << iOpponent
             << " cycle=" << WM->getCurrentCycle()
             << endl;
*/

      if ( WM->getTimeSinceLastCatch() < AFTER_CATCH_DELAY )
      {
        should = false;
      }
      else if( WM->isFreeKickUs() && WM->isBallInOwnPenaltyArea() )
      {
        should = false;
      }
      //I am a defender...
      else if ( fastestTeammateType == PT_DEFENDER_SWEEPER ||
                fastestTeammateType == PT_MIDFIELDER_WING     )
      {
        if ( posBall.getX() > 0 )
        {
          should = false;
        }
        else if ( not WM->isInDefenseArea( OBJECT_BALL )  &&
                  not WM->isDeadBallUs()                  &&
                  fabs( WM->getStrategicPosition().getY() - posBall.getY() ) > 7.0 )

        {
          should = false;
        }
      }

      if ( should )
      {
        // set the intercepting style:

        if ( ( fastestTeammateType == PT_MIDFIELDER_CENTER ||               // I am a midfielder, and
               fastestTeammateType == PT_MIDFIELDER_WING           ) &&
             ( secondFastestTeammateType == PT_DEFENDER_SWEEPER ||          // the second fastest is a defender
               secondFastestTeammateType == PT_DEFENDER_WING       )    )
        {
          *iType = INTERCEPTION_AS_SECOND;
        }

        else if ( fastestTeammateType == PT_ATTACKER ||                     // I am attacker
                  fastestTeammateType == PT_ATTACKER_WING )
        {
          *iType = INTERCEPTION_AS_FORWARD;
        }
        else if ( fastestTeammateType == PT_DEFENDER_SWEEPER ||             // I am a defender
                  fastestTeammateType == PT_DEFENDER_WING       )
        {
          *iType = INTERCEPTION_AS_DEFENDER;
        }
        else if ( fastestTeammateType == PT_MIDFIELDER_CENTER ||            // I am a midfielder
                  fastestTeammateType == PT_MIDFIELDER_WING )
        {
          *iType = INTERCEPTION_AS_FORWARD;
        }
      }
    }
    // end for me as the fastest teammate;
    // consider intercepting as a second fastest teammate
    else if ( secondFastestTeammate == WM->getAgentObjectType() &&          // I am the second fastest teammate
              fastestTeammate != fastestPlayer          &&                  // the opponent is faster, anyway
              WM->getAgentTiredNess() != TIREDNESS_BAD  &&                  // I am not exhausted
              not WM->isDeadBallUs()                                )       // leaving dead ball to the fastest teammate

    {
/*
      if ( WM->getPlayerNumber() <= 3 )
        cout << "...#" << WM->getPlayerNumber()
             << " I am the SECOND fastest "
             << " iFast=" << iFastest  << " #" << fastestTeammate - 60
             << " iTmm=" << iTeammate1
             << " iOpp=" << iOpponent
             << " cycle=" << WM->getCurrentCycle()
             << endl;
*/
      if ( SoccerTypes::isGoalie( fastestTeammate) )
      {
        should = true;
        *iType = INTERCEPTION_AS_DEFENDER;
      }

      else if( ( secondFastestTeammateType == PT_MIDFIELDER_CENTER ||        // I am a midfielder, and
                 secondFastestTeammateType == PT_MIDFIELDER_WING      ) &&
               ( fastestTeammateType == PT_DEFENDER_SWEEPER  ||              // the fastest teammate is defender
                 fastestTeammateType == PT_DEFENDER_WING         )        )
      {
        should = true;
        *iType = INTERCEPTION_AS_SECOND;
      }

      else if ( ( secondFastestTeammateType == PT_DEFENDER_WING   ||         // I am a defender, and
                  secondFastestTeammateType == PT_DEFENDER_SWEEPER   ) &&
                  iTeammate1 < 10                                      &&    // close enough to
                ( fastestTeammateType == PT_MIDFIELDER_CENTER ||             // the fastest teammate who is a midfield
                  fastestTeammateType == PT_MIDFIELDER_WING          )   )
      {
         should = true;
         *iType = INTERCEPTION_AS_DEFENDER;
      }
    } // end of me as second

    // override all above if I am receiving a pass
    if ( not should                                  &&
         WM->multiCycleAction == ACT_INTERCEPT_PASS  &&
         not WM->isDeadBallUs()                          )
    {
      // Although the agent does not belive that he should intercept, he hears a message
      // from the ball-possesing teammate that he is supposed to receive the pass now.
      // In this case, agent ignores everything above and attempts to start ball interception.
/*
      cout << "&&&& #" << WM->getPlayerNumber()
           << ": Intercepting in spite of all doubts." << endl;
*/
      *iType = INTERCEPTION_AS_SECOND;          // interception type doesn't matter in this case
      should = true;
    }

    // override all above to resolve a stalemate in the a ball-holding situation
    if ( not should                                         &&
         fastestPlayer == fastestTeammate                   &&
         WM->getCyclesBallStandingStill() > 15              &&
         secondFastestTeammate == WM->getAgentObjectType()  &&
         not WM->isDeadBallUs()                             &&
         WM->getCurrentCycle() > 5                          &&
         WM->getCurrentCycle() < 6000                           )
    {

      //cout << "&&&& #" << WM->getPlayerNumber()
      //     << ": Intercepting ball held by a teammate for too long. Cycle="<< WM->getCurrentCycle() << endl;

      *iType = INTERCEPTION_AS_SECOND;          // interception type doesn't matter in this case
      should = true;
    }

  }
  else
  {
    // ball kick us case; do nothing
  }

  return should;
}



   
 
// An alternative simplified method to incorporate optimization, and
// complement it whenever it returns unsatisfactory result.

ActionT Player::determineActionWithBall2()
{
  ActionT action = ACT_ILLEGAL;

  // general objects
  ObjectT     closestOpponent;
  ObjectT     agentObjectType = WM->getAgentObjectType();
  double      closestOpponentDistance = 0;

  // general positios
  VecPosition posBall = WM->getBallPos();
  VecPosition bestScoringPoint;

  
  if ( WM->isDeadBallUs() )
  {
     if(WM->getTimeSinceLastCatch() < PS->getCyclesCatchWait() ||
        WM->isGoalKickUs() )
     {
        action = ACT_MOVE_TO_DEAD_BALL_POSITION;
     }

     // for right kicking the agent must go to the right position and turn
     if ( WM->getAgentGlobalPosition().getDistanceTo( WM->getPosForDeadBall() ) > 0.5 ||
          WM->getRelativeAngle(OBJECT_BALL, true) > 15.0 )
     {
       if (WM->getAgentViewAngle() !=VA_WIDE)
       {
         ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_WIDE, VQ_HIGH ) );
       }
       action = ACT_MOVE_TO_DEAD_BALL_POSITION;
     }
  } // end of deadBallUs


  //*** shooting ***
  if ( action == ACT_ILLEGAL &&                // action has not been selected yet
       shouldIScore( &bestScoringPoint ) )
  {
    action = ACT_KICK_TO;
    pointToPass = bestScoringPoint;
  }

  
  if ( action == ACT_ILLEGAL )                 // action has not been selected yet
       
  {
    // *** optimization  *** try to find optimal kick ball parameters

    int numOfChoices;            // number of not too bad alternative options available

    // intitialize the grid for for decision making with the ball and
    // decide if the optimization method is applicable
    if ( initGrid( &numOfChoices ) )              
    {
      // try to find the optimal solution using multiple criteria
      if ( optimizeActionWithTheBall( numOfChoices ) )
      {
        // implement the optimal solution
        action =  ACT_KICK_TO_OPTIMIZED;
      }
    }
  }


  // try to trick the opponent goalie
  if ( action == ACT_ILLEGAL )
  {
    if ( shouldITrickOpponent( WM->getOppGoalieType() ) )
    {
      action = ACT_TRICK;
      teammateToPass = closestOpponent;

      //cout << "#" << WM->getPlayerNumber()
      //   << " is tricking the opponent goalie #" << SoccerTypes::getOpponentNumber( closestOpponent )
      //   << ". (case 0) cycle = " << WM->getCurrentCycle() << endl;

    }
  }

  if ( action == ACT_ILLEGAL  &&
       WM->multiCycleAction == ACT_PASS_ON_REQUEST )
  {
      currentAction = ACT_DIRECT_PASS;
  }

      
  if ( action != ACT_ILLEGAL )
    return action;
//<==============



  // *** initialization for dribbling  ***

  int i;  //an index for loops

  int dIndex = 6; //number of possible directions for diff purposes. Temporary here, must be defined somewere else

  // the list of general directions and angles
  AngDeg      widestForwardDirection = WM->getBestDirForDribble(0.0, 105.0);
  AngDeg      widestStrightForwardDirection = WM->getBestDirForDribble(0.0, 60.0);
  AngDeg      widestBackDirection = WM->getBestDirForDribble((VecPosition(0.0,0.0)-posBall).getDirection(), 60.0);
  AngDeg      towardsGoalDirection = (WM->getPosOpponentGoal()-posBall).getDirection();
  AngDeg      towardsPenaltyAreaDirection = (VecPosition(PITCH_LENGTH/2.0-11.0, 0.0)-posBall).getDirection();


  // directions for ordering
  AngDeg     directions[dIndex];
  for( i=0; i<dIndex; i++)
    directions[i] = 1000.0;

  AngDeg      toTurnDirection = 1000.0;
  AngDeg      leftStickRelDir = (VecPosition(PITCH_LENGTH/2.0, -8.5) - posBall).getDirection();
  AngDeg      rightStickRelDir = (VecPosition(PITCH_LENGTH/2.0, 8.5) - posBall).getDirection();

  towardsGoalDirection = WM->getBestDirForDribble(towardsGoalDirection, 60.0);

  towardsPenaltyAreaDirection = WM->getBestDirForDribble(towardsPenaltyAreaDirection, 60.0);
  leftStickRelDir = WM->getBestDirForDribble(leftStickRelDir, 60.0);
  rightStickRelDir = WM->getBestDirForDribble(rightStickRelDir, 60.0);

  if ( WM->isInDefenseArea( OBJECT_BALL ) )
  {
    //directions
    directions[0] = widestStrightForwardDirection;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    directions[4] = widestBackDirection;
    if (posBall.getY() > 0.0)
      directions[5] = (VecPosition(0.0, PITCH_WIDTH*0.5) - posBall).getDirection();
    else
      directions[5] = (VecPosition(0.0, -PITCH_WIDTH*0.5) - posBall).getDirection();
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);

    toTurnDirection = 0.0;

  }
  else if (WM->isInMiddleArea(OBJECT_BALL))
  {
     //directions
    directions[0] = widestStrightForwardDirection;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    toTurnDirection = 0.0;
    directions[4] = widestBackDirection;
    if(posBall.getX() < 0.0)
    {
      if (posBall.getY() > 0.0)
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();
      else
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();

    }
    else
    {
      if (posBall.getY() > 0.0)
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();
      else
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();
    }
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);

    toTurnDirection = 0.0;
  }
  else if (WM->isBallWithinRegion(REGION_4a) )
  {
     //directions

    directions[0] = towardsGoalDirection;
    directions[2] = widestStrightForwardDirection;
    directions[1] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    toTurnDirection = 0.0;
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);
  }
  else if (WM->isBallWithinRegion(REGION_4b) )
  {
    directions[0] = towardsGoalDirection;
    directions[2] = widestStrightForwardDirection;
    directions[1] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    toTurnDirection = towardsGoalDirection;
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);
  }
  else if (WM->isBallWithinRegion(REGION_5a) )
  {
    //directions[0] = towardsPenaltyAreaDirection;
    directions[0] = leftStickRelDir;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = (VecPosition(PITCH_LENGTH/2.0-25.0, 0.0)-posBall).getDirection();
    directions[3] = WM->getBestDirForDribble(directions[2], 20.0);
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    directions[5] = directions[0];

    toTurnDirection = towardsPenaltyAreaDirection;
   // widestBackDirection = WM->getBestDirForDribble(179.0, 20.0);
  }
  else if (WM->isBallWithinRegion(REGION_5b) )
  {
   // directions[0] = towardsPenaltyAreaDirection;
    directions[0] = rightStickRelDir;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = (VecPosition(PITCH_LENGTH/2.0-25.0, 0.0)-posBall).getDirection();
    directions[3] = WM->getBestDirForDribble(directions[2], 20.0);
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    //if (posBall.getX() < PITCH_LENGTH/2.0 - 7.0)
     // directions[3] = widestForwardDirection;
    toTurnDirection = towardsPenaltyAreaDirection;
    //widestBackDirection = WM->getBestDirForDribble(179.0, 20.0);
    directions[5] = directions[0];
  }
  else if (WM->isBallWithinRegion(REGION_6) )
  {
    directions[2] = widestStrightForwardDirection;
    directions[0] = towardsGoalDirection;
    directions[1] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    toTurnDirection = 0.0;
    if (posBall.getY() > 0.0)
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();
      else
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();

    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);

  }
   else if (WM->isBallWithinRegion(REGION_6a) )
  {
    directions[0] = towardsPenaltyAreaDirection;
    directions[1] = towardsGoalDirection;
    directions[4] = widestBackDirection;
    directions[5] = directions[0];

    toTurnDirection = towardsPenaltyAreaDirection;
  }

  //  *** end of initialization  ***
  
  // try to dribble fast
  if ( action == ACT_ILLEGAL )
  {
    for( i=0; i < 3; i++ ) // empiric number here, must be defined somewhere else
    {
      if ( directions[i] < 900.0 && WM->canItoDribble( directions[i], DRIBBLE_FAST ) )
      {
        cout << "#" << WM->getPlayerNumber() <<" dribble fast, cycle = " << WM->getCurrentCycle() << endl;
        action = ACT_DRIBBLE;
        dribbleDirection = directions[i];
        pointToPass = posBall;
        dribbleType = DRIBBLE_FAST;
        break;
      }
    }
  }

    
  // dribble slow
  if ( action == ACT_ILLEGAL ) // action wasn't chosen before
  {
    for(i=0; i < 3; i++) // empiric number here, must be defined somewhere else
    {
      if(directions[i] < 900.0 && WM->canItoDribble( directions[i], DRIBBLE_SLOW))
      {
        cout<< "#" << WM->getPlayerNumber() <<" dribble slow, cycle = " << WM->getCurrentCycle() << endl;
        action = ACT_DRIBBLE;
        dribbleDirection = directions[i];
        pointToPass = posBall;
        dribbleType = DRIBBLE_SLOW;
        break;
      }
    }
  }
  
  // try to pass the ball to closely located free teammate
  if ( action == ACT_ILLEGAL &&
       WM->isItSafeToPassClose( &teammateToPass ) )
  {
    action = ACT_DIRECT_PASS;

    cout << "#" << WM->getPlayerNumber()
         << ": short direct pass to #" << SoccerTypes::getTeammateNumber( teammateToPass )
         << ".  cycle = " << WM->getCurrentCycle() << endl;

  }

  //if need to turn
  if ( action == ACT_ILLEGAL ) // action wasn't chosen before
  {
   if( WM->canITurn(toTurnDirection) )
    {
      cout<< "#" << WM->getPlayerNumber() <<" turn "<< endl;
      action = ACT_TURN_WITH_BALL;
      dribbleDirection = toTurnDirection;
    }
  }

    closestOpponent = WM->getClosestInSetTo( OBJECT_SET_OPPONENTS, agentObjectType,
                                                            &closestOpponentDistance );
  // try to trick closest opponent
  if ( action == ACT_ILLEGAL &&
       closestOpponent != OBJECT_ILLEGAL &&
       shouldITrickOpponent( closestOpponent ) )
  {
    action = ACT_TRICK;
    teammateToPass = closestOpponent;

    cout << "#" << WM->getPlayerNumber()
         << " is tricking opponent #" << SoccerTypes::getOpponentNumber( closestOpponent )
         << ".  cycle = " << WM->getCurrentCycle() << endl;
  }

  // try to hide ball
  if ( action == ACT_ILLEGAL  &&
       closestOpponent != OBJECT_ILLEGAL )
  {
    cout << "#" << WM->getPlayerNumber() << " hiding ball.  cycle = " << WM->getCurrentCycle() << endl;
    action = ACT_HIDE_BALL;
    teammateToPass = closestOpponent;
  }

  return action;

} // determineActionWithBall2




/*! This method prints out statistics collected for each player
    It is standalone just for making changes easier           */
void Player::printStatistics( )
{
/*
printf("Shutting down player ============== %d\n", WM->getPlayerNumber() );

  printf("   Number of holes: %d (%f)\n", WM->iNrHoles,
                         ((double)WM->iNrHoles/WM->getCurrentCycle())*100 );

  printf("   Teammates seen: %d (%f)\n", WM->iNrTeammatesSeen,
                         ((double)WM->iNrTeammatesSeen/WM->getCurrentCycle()) );
  printf("   Opponents seen: %d (%f)\n", WM->iNrOpponentsSeen,
                         ((double)WM->iNrOpponentsSeen/WM->getCurrentCycle()) );
  printf("   Ball seen: %d (%f)\n", WM->iNrBallSeen,
                         ((double)WM->iNrBallSeen/WM->getCurrentCycle()) );

  printf("   Ball in our possesion: %d (%f)\n", WM->iNrBallPossesionUs,
                         ((double)WM->iNrBallPossesionUs/WM->getCurrentCycle()) );
  printf("   Ball on the opponent side: %d (%f)\n", WM->iNrBallSideThem,
                         ((double)WM->iNrBallSideThem/WM->getCurrentCycle()) );  
  printf("   Ball average X: %f\n", WM->dBallAvgXcoordinate/(double)WM->getCurrentCycle() );

  printf("   Say-ball messages sent: %d (%f)\n", WM->iNrSayBallSent,
                         ((double)WM->iNrSayBallSent/WM->getCurrentCycle()) );
  printf("   Say-ball messages heard: %d (%f)\n", WM->iNrSayBallHeard,
                         ((double)WM->iNrSayBallHeard/WM->getCurrentCycle()) );
  printf("   Say-ball messages used: %d (%f)\n", WM->iNrSayBallUsed,
                         ((double)WM->iNrSayBallUsed/WM->getCurrentCycle()) );
  printf("   Teammate positions updated: %d (%f)\n", WM->iNrTeammatesUpdated,
                         ((double)WM->iNrTeammatesUpdated/WM->getCurrentCycle()) );
*/
  printf("====    SFUnleashed'03 player  #%d", WM->getPlayerNumber() );
  printf("  is saying good bye.   ====\n");
  printf("  Developed using **basic 2002 UvA Trilearn** as a prototype \n");
  printf("  by Vadym Kyrylov (Vadim Kirillov), project director\n" );
  printf("               Serguei Razykov,      team leader\n " );
  printf("               David Brokenshire,    co-developer, and \n" );
  printf("               Eddie Hou,            research contributor\n" );
  printf("  (c) 2003  \n" );
}


//!  This method returns ball confidence used in sending audial
//   message. In most cases we want the updated ball state send.
//   However, in some cases the quality of this information cold be
//   poor, but the messages is still sent. Confidence = 0.97 means that
//   it is poor indeed and should not be completely used by the receiving player.
double Player::getConfidenceSayBall()
{
  double myconfidence;

  if ( WM->iSayBallMsgCode != WM->MSG_OPPONENT_CLOSE )
  {
    if ( WM->getAgentGlobalPosition().getDistanceTo( WM->getBallPos() ) > 15.0 )
      myconfidence = 0.97;
    else
      myconfidence = WM->getConfidence( OBJECT_BALL );  // default value
  }
  else
  {
    // use the opponent player intead the ball 
    ObjectT opp = SoccerTypes::getOpponentObjectFromIndex( WM->iPlayerInvolved );
    myconfidence = WM->getConfidence( opp );
  }
  
  return myconfidence;
}



/*! This method generates a SayBall message.
    Player sends this message if sufficient time expired after
    the previous message (default for this limit is zero)
    also, player must have high enough confidence about the ball.
    bNeedToSayBall and bRetransmitSayBall flags can be raised by
    the world model update methods.                               */
void Player::sendSayBallMsg()
{
      // Override retransmitting whenever new info is avalable to say.
      if ( WM->bNeedToSayBall ) WM->bRetransmitSayBall = false;
            
      if ( WM->bNeedToSayBall or WM->bRetransmitSayBall )
      if ( WM->getCurrentCycle() > 0 )  // shut up until the kick-off
      {        
        double confidence = getConfidenceSayBall();
        iSayBallCount++;
        if ( iSayBallCount >= WM->iSayBallInterval &&   // I am not sending it too frequently AND
             confidence > 0.94 )                        // am sending if only I have fairly good visual info
         {          
/*          
          cout <<"**** say ball = " << WM->getPlayerNumber()
                <<" time = " << WM->getCurrentCycle()
                <<" conf = " << WM->getConfidence( OBJECT_BALL )
                << " iNrSayBallSent=" << WM->iNrSayBallSent
                << endl;
*/
          // input parameter values
          // must be assigned while raising bNeedToSayBall flag.
          // iNrCyclesToComplete is used not in all message types
          if ( WM->iPlayerInvolved > 0     &&
               WM->iPlayerInvolved < 12  ) //  &&
            //   WM->getCurrentCycle() > WM->iCylcleLastSayBall )         // prevent from repeating in same cycle
            ACT->sendCommandDirect( sayBallStatus( WM->iSayBallMsgCode,
                                                  WM->iPlayerInvolved,
                                                  WM->iNrCyclesToComplete,
                                                  confidence,
                                                  WM->bSayBallLowStamina,
                                                  WM->bRetransmitSayBall
                                  )
          );

          iSayBallCount = 0;
          WM->iCylcleLastSayBall = WM->getCurrentCycle();
          WM->iNrSayBallSent++;     // accumulate player statistics

        }
      }
} // sendSayBallMsg



/*! This method is used to communicate the status of the ball, that is its
    position and velocity, and the actions executed with it.

    Encoding conventions:
    ---------------------
    Each byte of the message can contain take one of total 75 character
    values as specified in WorldModel::getCharToInt .
    Of them, only 64 are used to store numerical data (i.e. one char for 6 bits).

    Byte 0: Message type in 11*Action_Type format (1..66, 72, 73)
         1..3: x,y-coordinates (8 bits each) and retransmit flag (1 bit)
         4..6: Vx, Vy (8 bits each)
         7..8: xy 6-bit coordinates of an object, such as the pass interception point or TA
         8..9: confidence factor (2 bits) and time to complete the action (4 bits)

     For message/action type conventions, see action codes in Player.h
     For each type, 11 numbers are allocated for the respective players involved into
     the action with the ball. This player can be either the message sender, or the pass
     receiver, or the opponent who is the fastest to the ball.

     Retransmitted messages:
          retransmit flag is encoded in x-coordinate (Byte1)
          xy-coordinates are extrapolated for the time of retransmitting the message
          time to complete the pass reduced by one cycle
          confidence is reduced accordingly

     Byte-9 conventions:
       The encoded value (0..63) is 4*N+C, where
       N = number of cycles to complete the action (0..15)
       C = 3-percent reduction of the confidence factor (0..3)
    */
SoccerCommand Player::sayBallStatus( int ballActionType,
                                     int playerID,        // me, pass receiver, or the opponent
                                     int nr_cyclces,      // time parameter
                                     double confidence,
                                     bool lowSta,         // low stamina flag
                                     bool retransmit )
{
  char    strMsg[MAX_SAY_MSG];
  char    MsgType;
  int     MsgNumber;

  if ( ballActionType == WM->MSG_NOMESSAGE )
  return SoccerCommand( CMD_ILLEGAL, "MsgType=0" );      // do nothing (for robustness)


  if ( ballActionType == WM->MSG_I_DRIBBLE )
  { // get the info about the target area  vpSayBallPoint
    // and include it into the message (not implemented yet)
    ;
  }

  // Make message type
  MsgNumber = 11 * ( ballActionType - 1 ) + playerID;
  MsgType = WM->getIntToChar( MsgNumber );

  // anticipated time to execute the action with he ball
  nr_cyclces = max( 0, nr_cyclces );
  nr_cyclces = min( 15, nr_cyclces );

  VecPosition posBall, velBall;

  if ( ballActionType == WM->MSG_OPPONENT_CLOSE )
  {
    ObjectT opp = SoccerTypes::getOpponentObjectFromIndex( playerID - 1 );
    posBall = WM->getGlobalPosition( opp );
    velBall = WM->getGlobalVelocity( opp );
  }  
  else 
  {
    // use ball state vector in all the rest cases (even if the message is retransmitted) 
    posBall = WM->getGlobalPosition( OBJECT_BALL );
    velBall = WM->getGlobalVelocity( OBJECT_BALL );
  }
    

/**** test
  posBall.setVecPosition( 9.4, -9.4 );
  velBall.setVecPosition( -0.9, 0.9 );
  WM->vpSayBallPoint.setVecPosition( 14.0, -14.0 );
  confidence = 0.971;
  nr_cyclces = 10;
  retransmit = true;
  lowSta = false;
****/

  // conf = 0..3 corresponds to confidence = 1.00..0.97
  // 0.97 is the critical value meaning "0.97 or less"; do not
  // utilize coordinate info when confidence = 0.97
  confidence = min( 1.0, confidence );  
  confidence = max( 0.0, confidence );
  int conf = (int)( (1.0 - confidence)*100.0 );
  conf = max ( conf, 0 );
  conf = min ( conf, 3 );

  static double maxXY = 50.0;
  static double maxV = 3.0;

  // map ball position from (-maxXY, maxXY) to 0..255
  int x = WM->convertDblXYtoIntScale( posBall.getX(), maxXY, 255 );
  int y = WM->convertDblXYtoIntScale( posBall.getY(), maxXY, 255 );

  // split x,y into three chunks to make virtual 8-bit representation
  int chunk1 = WM->convert2in3( x, y, 1);         // 4 upper 'bits' of x
  int chunk2 = WM->convert2in3( x, y, 2);         // 4 lower 'bits' of x and 2 lower 'bits' of y
  int chunk3 = WM->convert2in3( x, y, 3);         // 6 lower 'bits' of y

  if ( retransmit )
  {
    chunk1 = chunk1 + 32; // use 6-th virtual bit
    // if message is retransmitted, decrement remaining time and confidence
    nr_cyclces = max( 0, nr_cyclces - 1 );
    conf = min( 3, conf + 1 );
  }

  if ( lowSta )
  chunk1 = chunk1 + 16; // use 5-th virtual bit

  // map ball velocity from (-maxV, maxV) to 0..255
  int vx = WM->convertDblXYtoIntScale( velBall.getX(), maxV, 255 );
  int vy = WM->convertDblXYtoIntScale( velBall.getY(), maxV, 255 );

  // split vx, vy into three chunks to make virtual 8-bit representation
  int chunk4 = WM->convert2in3( vx, vy, 1);         // 4 upper 'bits' of vx
  int chunk5 = WM->convert2in3( vx, vy, 2);         // 4 lower 'bits' of vx and 2 lower 'bits' of vy
  int chunk6 = WM->convert2in3( vx, vy, 3);         // 6 lower 'bits' of vy

  // map point of attention position from (-maxXY, maxXY) to 0..63
  // ( vpSayBallPoint must be assigned elsewhere )
  int xTA = WM->convertDblXYtoIntScale( WM->vpSayBallPoint.getX(), maxXY, 63 );
  int yTA = WM->convertDblXYtoIntScale( WM->vpSayBallPoint.getY(), maxXY, 63 );

  sprintf(  strMsg, "%c%c%c%c%c%c%c%c%c%c",
            MsgType ,
            WM->getIntToChar( chunk1 ),
            WM->getIntToChar( chunk2 ),
            WM->getIntToChar( chunk3 ),
            WM->getIntToChar( chunk4 ),
            WM->getIntToChar( chunk5 ),
            WM->getIntToChar( chunk6 ),
            WM->getIntToChar( xTA ),
            WM->getIntToChar( yTA ),
            WM->getIntToChar( nr_cyclces*4 + conf )
  );


//  string repeat = !retransmit ? "" : " repeatedly";

//  if ( lowSta ) cout << "low stamina plr = " << playerID << endl;

//  if ( WM->getPlayerNumber() == 1 )
//  if ( ballActionType == WM->MSG_PASSING_TO && !retransmit )
//  cout << "+++ say #" << WM->getPlayerNumber()
//      << ": " << strMsg
//      << " "<< WM->getMsgCodeFromIdx ( ballActionType )
//      << " code = " << ballActionType
//      << " plr="<< playerID
//      << " retr="<< retransmit
//      << " lowSta="<< lowSta
//      << endl
//      << " xB="<< posBall.getX()
//      << " yB="<< posBall.getY()
//      << " vx="<< velBall.getX()
//      << " vy="<< velBall.getY()
//      << endl
//      << " mypos="<< WM->vpSayBallPoint
//      << "++ xTA="<< WM->vpSayBallPoint.getX()
//      << " yTA="<< WM->vpSayBallPoint.getY()
//      << " cycles="<< nr_cyclces
//      << " cnf="<< confidence
//      <<" time = " << WM->getCurrentCycle()
//      << repeat
//      << endl;

      
  return communicate( strMsg );

} // sayBallStatus



//!  This method makes a decision for the golie agent whether to request
//   the ball owner to pass the ball back to him.
//   Agent requests for passing if only there is a good chance for the opponents to
//   successfully continue the attack on our goal.
//    \param pointToPass is the point where the ball is requested to be passed
//    \param iCycles is the estimated minimal number of cycles for completing the pass
bool Player::shouldGoalieRequest( VecPosition *pointToPass, int *iCycles )
{
  const double maxDistance = 17.0;
  const double wideSectorWidth = 17.0;

  bool should = false;

  if ( WM->getBallPossesion() != POSSESION_THEM )
  {
    VecPosition ballpsn = WM->getGlobalPosition( OBJECT_BALL);
    double distToBall = WM->getAgentGlobalPosition().getDistanceTo( ballpsn );

    if ( distToBall < maxDistance )     // everything happens close to the goal
    {
      // check if the teammate is indeed fastest to the ball
      int iCyclesToIntercept;
      ObjectT tmm = WM->getFastestInSetTo( OBJECT_SET_PLAYERS, OBJECT_BALL, &iCyclesToIntercept );

      if ( SoccerTypes::isTeammate( tmm ) &&
           WM->getAgentObjectType() != tmm    &&        // not myself
           iCyclesToIntercept < 3 )
      {
        // the teammate is near the ball (otherwise, no sense to request for passing)
        // now count the opponents in the ball neighbourhood ( 6m x 8m )
        Rectangle ballRect = Rectangle( ballpsn - VecPosition ( 3.0, 4.0 ),
                                        ballpsn + VecPosition ( 3.0, 4.0 ) );
        int nrofOppClose = WM->getNrInSetInRectangle( OBJECT_SET_OPPONENTS, &ballRect );
        if ( nrofOppClose > 0 )
        {
          // at least one opponent is there...
          VecPosition goalpsn = WM->getPosOwnGoal();
          // count the opponents in the wide sector centered at ball and pointed at goalpsn
          int nrOfOppWide = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                                  wideSectorWidth/(2.0*distToBall),
                                                            ballpsn, goalpsn );
          if ( nrOfOppWide > 0 )
          {
            // at least one opponent is very close to the goal... tough...
            VecPosition mypsn = WM->getAgentGlobalPosition();
            // count the opponents in a ~15 degree sector centered at ball and pointed at mypsn
            int nrOfOppNarrow = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS, 0.15,
                                                                  ballpsn, mypsn );
            if ( nrOfOppNarrow == 0 )
            {
              // it is clear ...
              // so, the situation could be saved by passing the ball to me the goalie
              should = true;
              *pointToPass = mypsn;
              int iCyclesBall = WM->getNrCyclesBallToMove( ballpsn, mypsn, PS->getSlowPassEndSpeed() );
              *iCycles = iCyclesBall;
              //cout << "-- shouldGoalieRequest()=true.  Time = " << WM->getCurrentCycle() << endl;
            }
          }
        }
      }
    }
  }
  return should;
} // shouldGoalieRequest



//!  This method makes a decision for the agent whether to request
//   the ball owner to pass the ball to the designated point.
//   Agent requests for passing if only he has a good chance to
//   attack the opponent goal.
//    \param pointToPass is the point where the ball is requested to be passed
//    \param iCycles is the estimated minimal number of cycles for completing the pass

bool Player::shouldIRequestForPassing ( )
{
  return false;                                   // DISABLED

  const double maxDistance = 25.0;
  const double wideSectorWidth = 16.0;
  double nrOfOppNarrow, nrOfOppWide;
  VecPosition mypoint;
  int iCyclesMe;

  bool should = false;

  if ( SoccerTypes::isGoalie( WM->getAgentObjectType() ) )
  {
//    should = shouldGoalieRequest( &mypoint, &iCyclesMe);
  }


  if ( !should && WM->getBallPossesion() == POSSESION_US )
  {
    // this doesn't apply to defenders (better change to isDefender() )
    if ( WM->getPlayerNumber() < 6 ) return false;
//<-----

    // and doesn't apply to exhausted agent either
    if ( WM->getAgentTiredNess() == TIREDNESS_BAD  ||
         WM->getAgentTiredNess() == TIREDNESS_VERY_BAD )
      return false;
//<-----

    VecPosition ballpos = WM->getGlobalPosition( OBJECT_BALL) ;
    double distToBall = WM->getAgentGlobalPosition().getDistanceTo( ballpos );

    if ( distToBall < maxDistance  &&  distToBall > 0.4*maxDistance )
    {
      // I am not far away and not very close to the ball
      VecPosition mypsn = WM->getAgentGlobalPosition();
      VecPosition goalpsn = WM->getPosOpponentGoal();
      double myDistToGoal = mypsn.getDistanceTo( goalpsn );

      if ( myDistToGoal < maxDistance*2.5 )
      {
        // check if the ball-possessing teammate might have dificulties
        bool hasproblem = false;
        double hisDistToGoal = ballpos.getDistanceTo( goalpsn );

        if ( hisDistToGoal > maxDistance/2.0 )
        {
          if ( hisDistToGoal > maxDistance )
            hisDistToGoal = maxDistance;
          // he is not very close to the opponent goal yet...

          // check if the teammate is indeed fastest to the ball
          int iCyclesToIntercept;
          ObjectT tmm = WM->getFastestInSetTo( OBJECT_SET_PLAYERS, OBJECT_BALL, &iCyclesToIntercept );

          if ( SoccerTypes::isTeammate( tmm ) &&
               WM->getAgentObjectType() != tmm    &&        // not myself
               iCyclesToIntercept < 3 )                     // the teammate is near the ball
          {
            // am I significantly more advanced to the opponent goal than him ?
            if ( WM->getGlobalPosition( tmm ).getDistanceTo( goalpsn )
                                     - myDistToGoal > 0.33*maxDistance )
            {
              hasproblem = true;
            }
            else
            {
              // check for more serious problems:
              // count the opponents in the narrow sector centered at ball and pointed at goalpsn
              nrOfOppNarrow = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                                    wideSectorWidth/(4.0*hisDistToGoal),
                                                     ballpos, goalpsn );
              if ( nrOfOppNarrow > 1.0 )
                // the teammate appears to have little chance of success with scoring the goal now
                hasproblem = true;
              else
              {
                // count the opponents in the wide sector centered at ball and pointed at goalpsn
                nrOfOppWide = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                                    wideSectorWidth/(2.0*hisDistToGoal),
                                                          ballpos, goalpsn );
                if ( nrOfOppWide > 2.0 )
                  // the teammate appears to have little chance of success with scoring the goal, anyway
                  hasproblem = true;
              }
            }
          }
        }

//        hasproblem = true;  // A PATCH for testing

        if ( hasproblem )
        {
          // check if there is reasonably clear space between me and the opponent goal:
          if ( myDistToGoal > maxDistance )
            myDistToGoal = maxDistance;
          // count the opponents in the narrow sector centered at mypsn and pointed at goalpsn
          nrOfOppNarrow = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                            wideSectorWidth/(4.0*myDistToGoal), mypsn, goalpsn );
          // count the opponents in a ~30-degree sector centered at mypsn and pointed at goalpsn
          nrOfOppWide = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                            wideSectorWidth/(2.0*myDistToGoal), mypsn, goalpsn );
          double nrOfOpponents = nrOfOppNarrow + 0.5 * nrOfOppWide;

          if ( nrOfOpponents < 2.6 )
          {
            // there is likely the opponent goalie only... This is a golden opportunity to score.
            // Now find the first good pass-receiving point by moving along the line to the goal
            // starting from fartherstPoint to closestPoint.

            VecPosition fartherstPoint = goalpsn * ( 1.0 - 0.35 ) + WM->getAgentGlobalPosition() * 0.35;
            VecPosition closestPoint   = goalpsn * ( 1.0 - 0.90 ) + WM->getAgentGlobalPosition() * 0.90;
            
            mypoint = WM->getIntercepionPointOnTheLine( WM->getGlobalPosition( OBJECT_BALL ), 
                                                    4, closestPoint, fartherstPoint, 1, &iCyclesMe, &should );
          }
        } // hasproblem
      }
    }

  }

  if ( should )
  {
    // prepare to send the message requesting to pass the ball
    WM->bNeedToSayBall      = true;
    WM->iSayBallMsgCode     = WM->MSG_PASS_ME_BALL;
    WM->vpSayBallPoint      = mypoint;
    WM->iNrCyclesToComplete = iCyclesMe;
    WM->iPlayerInvolved     = WM->getPlayerNumber();
/*
        cout <<"#" << WM->getPlayerNumber() << ": PASS_ME_BALL "
//          << " to "  << WM->vpSayBallPoint
//          << "   ax = " << WM->getAgentGlobalPosition().getX()
//          << " ay = " << WM->getAgentGlobalPosition().getY()
//          << endl
          << " time = " << WM->getCurrentCycle()
          << " cycles = " << WM->iNrCyclesToComplete
//          << " x = " << WM->vpSayBallPoint.getX()
//          << " y = " << WM->vpSayBallPoint.getY()
          << endl;
*/
  }

  return should;
}


//  This method returns true if it is worth trying to trick the opponent goalie
bool Player::shouldITrickGoalie()
{
  int iOpponent = 100;
  ObjectT fastestOpponent = WM->getFastestInSetTo( OBJECT_SET_OPPONENTS, OBJECT_BALL, &iOpponent );
  if ( iOpponent > 5 )
    return shouldITrickOpponent( WM->getOppGoalieType() );
  else
    return false;
}


//  This method works with trickOppMultyCycle.
//  The agent should try to trick the opponent "o" if he is close, but not very close and 
//  the relative angle between him and the ball is less than 90 degrees.
//  author Vadim K.
bool Player::shouldITrickOpponent( ObjectT o )
{
  bool trick = true;

  if ( !WM->isConfidenceGood( OBJECT_BALL ) )
  {
    trick = false;
  }
  else if ( !WM->isBallKickable() )
  {
    trick = false;
  }
  else if ( WM->getConfidence( o ) < 0.90 )
  {
    trick = false;
  }
  else
  {
    AngDeg angOpp  = WM->getRelativeAngle( o, true );               // direction to opponent relative to agent body
    AngDeg angBall = WM->getRelativeAngle( OBJECT_BALL, true );     // direction to ball relative to agent body
    AngDeg angDiff = fabs( VecPosition::normalizeAngle( angOpp - angBall ) );

    double coeff = 2.0;
    if ( o == WM->getOppGoalieType() )
      coeff = 2.5;

    if ( angDiff > 90.0 )
    {
      // ball is blocked from the opponent, anyway
      trick = false;
    }
    else
    {  
      double dDist = WM->getGlobalPosition(o).getDistanceTo( WM->getBallPos() );

      if ( dDist >  PS->getOppCloseThreshold() * coeff  ||
           dDist < PS->getOppVeryCloseThreshold()        )
      {
        trick = false;
      }
      else
      {  
        AngDeg bodyDirDiff = VecPosition::normalizeAngle( WM->getAgentGlobalBodyAngle() -
                                                          WM->getGlobalBodyAngle( o )      );
        // agent and opponent body directions must be somewhat opposite-intersecting
        if ( fabs ( bodyDirDiff ) < 90.0 )         
          trick = false;
      }
    }  
  }  

  return trick;

}


//  Trick the opponent by misleading him about the direction of the ball is heading to.
//  This effect is achievd by kicking the ball close to the body clockwise if the opponent
//  approached from the right, and counter-clockwise, if the opponent approches from the left.
//  It is expected that agent eventually will see a suitable teammate to pass him the ball.
//  At least, the ball will likely be blocked from the opponent by the agent's body.
//    \param 'o' is the object type of the opponent who should be tricked.
LevelStateT   Player::trickOppMultyCycle(ObjectT o)
{
  LevelStateT lst;
  const int maxSteps = 6;

  AngDeg angOpp  = WM->getRelativeAngle( o, true );              // direction to opponent relative to agent body
  AngDeg angBall = WM->getRelativeAngle( OBJECT_BALL, true );    // direction to ball relative to agent body

  // the termination condition for this multi-cycle action
  if ( multiCycleDribbleStep > maxSteps )
  {
    multiCycleDribbleStep = 0;
    lst =  ACTIONS_UNSELECTED_FINISHED;
//    cout << "-- terminated after " << maxSteps << " steps " << endl;
  }
  // other conditions for exit from this method
  else if ( !WM->isConfidenceGood( OBJECT_BALL ) )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
//    cout << "-- terminated: too low ball confidence " << endl;
  }
  else if ( !WM->isBallKickable() )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
//    cout << "-- terminated: ball is not kickable " << endl;
  }
  else if ( WM->getConfidence( o ) < 0.90 )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
//    cout << "-- terminated: opponent confidence too low " << endl;
  }
  else if ( WM->getGlobalPosition(o).getDistanceTo( WM->getBallPos() ) < PS->getOppVeryCloseThreshold() )        
  {
//    cout << "-- terminated: opponent too close " << endl;
    lst = ACTIONS_UNSELECTED_FINISHED;
  }
  else if ( fabs( VecPosition::normalizeAngle( angOpp - angBall ) ) > 160.0 )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
//    cout << "#" << WM->getPlayerNumber()
//         << " -- terminated: opponent has been blocked from the ball"
//         << ", cycle=" << WM->getCurrentCycle()
//         << endl;

  }
  else 
  {
    // ***  try to trick him !!!  ***
    SoccerCommand soc;

    if ( multiCycleDribbleStep == 0 )
    {
      // on Step 0, determine the direction of turn and set the ball/body turn increment    
      AngDeg angDiff = VecPosition::normalizeAngle( angOpp - angBall );  // can be positive or negative
      double dDist = WM->getGlobalPosition(o).getDistanceTo( WM->getBallPos() );
      // assuming that the opponent approaches the ball with 10 m/s, dDist is roughly the number of cycles.
      // I want to turn the ball by at least 240 degrees during this time
      ballTurnIncrement = 240.0 / dDist;
      if ( ballTurnIncrement < 30.0 )
        ballTurnIncrement = 30.0;                     // however, I should not be too slow
      if ( ballTurnIncrement > 170.0 )
        ballTurnIncrement = 170.0;                     // neither should I be too fast

      if ( o != WM->getOppGoalieType() || dDist < 2.5 )
      {  
        // field opponent player OR any player very close
        if ( dDist > 5.0  &&  WM->getGlobalVelocity( o ).getMagnitude() > 0.7  )
        {      
          // turn the ball closer TO the oppponent (to create a dynamic error in his movement)
          ballTurnIncrement = ballTurnIncrement * sign( angDiff );
        }
        else  
        {
          // turn the ball FROM the opponent
          ballTurnIncrement = - ballTurnIncrement * sign( angDiff );
        }
      }     
      else
      {
        // special case for the opponent goalie located not very close
        VecPosition posGoalie = WM->getOpponentGoaliePos();
        VecPosition leftStick = VecPosition(PITCH_LENGTH/2.0, -6.5);
        VecPosition rightStick = VecPosition(PITCH_LENGTH/2.0, 6.5);
        VecPosition posBall = WM->getBallPos();
        AngDeg      goalieRelDir = (posGoalie-posBall).getDirection();
        AngDeg      leftStickRelDir = (leftStick - posBall).getDirection();
        AngDeg      rightStickRelDir = (rightStick - posBall).getDirection();
        AngDeg      angleR = fabs( VecPosition::normalizeAngle( rightStickRelDir - goalieRelDir ) );
        AngDeg      angleL = fabs( VecPosition::normalizeAngle( leftStickRelDir - goalieRelDir ) );

        if ( angleR < angleL )
        {
          // turn the ball to the RIGHT of the opponent goalie
          ballTurnIncrement = ballTurnIncrement;
        }
        else
        {
          // turn the ball to the LEFT of the opponent goalie
          ballTurnIncrement = - ballTurnIncrement;
        }
      }
    }
    // kick the ball close to body in the direction of the side where the opponent is approaching
    AngDeg angleToTurnBall = VecPosition::normalizeAngle( angBall + ballTurnIncrement );
    soc = kickBallCloseToBody( angleToTurnBall, 0.50 );        
                                                                     
                                                                     
    lst = ACTIONS_SELECTED;
/*
    cout << "#" << WM->getPlayerNumber()
         << ", tricking opponent #" << SoccerTypes::getOpponentNumber( o )
//         << ", increment=" << ballTurnIncrement
         << ", cycle=" << WM->getCurrentCycle()
         << ", step=" << multiCycleDribbleStep
//         << endl
//         << "bodyDirG=" << WM->getAgentGlobalBodyAngle()
         << ", ballDirG=" << ( WM->getGlobalPosition( OBJECT_BALL ) - WM->getAgentGlobalPosition() ).getDirection()
//         << ", ballDirR=" << angBall
//         << ", oppDirG=" << ( WM->getGlobalPosition( o ) - WM->getAgentGlobalPosition() ).getDirection()
//         << endl
//         << "AgBallDist=" << WM->getAgentGlobalPosition().getDistanceTo( WM->getBallPos() )
//         << ", oppDist=" << WM->getGlobalPosition(o).getDistanceTo( WM->getAgentGlobalPosition() )
//         << ", oppBallDist=" << WM->getGlobalPosition(o).getDistanceTo( WM->getBallPos() )
         << endl;
*/

    if (soc.commandType != CMD_ILLEGAL)
    {      
      ACT->putCommandInQueue( soc );
      if ( WM->isLastMessageSee() )
        ACT->putCommandInQueue( turnNeckDiffSide( soc) );
    }

    multiCycleDribbleStep = nextMultyCycleCount();
//    cout << "exiting with step count = " <<  multiCycleDribbleStep << endl;
  }

  if ( lst != ACTIONS_SELECTED )
  {
    multiCycleDribbleStep = 0;
  }
    
  return lst;

}


//  This method advances the multi-cycle counter, multiCycleDribbleStep.
//  It prevents from incrementing the counter more than one time in each cycle
int Player::nextMultyCycleCount()
{
  int count = multiCycleDribbleStep;
  int cycle = WM->getCurrentCycle();

  if ( cycle > cycleMultiCounterLastUpdated )
  {
    count ++;
    cycleMultiCounterLastUpdated = cycle;
  }

  return count;
}


///////////////////////////////////////////////////////////////////////////////////
//                     OPTIMIZATION  FUNCTIONS
///////////////////////////////////////////////////////////////////////////////////


/******************************************************************************/
/**********************  CLASS GRIDCELL   *************************************/
/******************************************************************************/
// This class contains info about cells of the grid used by the player for
// making decisions about actions with the ball. Each cell is a sector.
// The cell collection forms several concentric rings.


// dummy constructor (it doesn't compile without it for some reason)
GridCell::GridCell ( )
{
  isActive = true;
}



//! This method assigns values to cell geometry attributes.
//  we actually are interested in the cell center and its area,
//  without precise knowledge of its shape and borders;
//  however, in these calcualations we treat cells as sectors
//    \param agPos is the agent global position
//    \param agDir is the agent global body direction
//    \param ringRadius is the distance from the grid center to the cell center
//    \param ang is the direction from the grid center to the cell center
//    \param area is the cell area
void GridCell::initCell( VecPosition agPos, AngDeg agDir,
                         double ringRadius, AngDeg cellAng, double area )
{
  isActive = true;
  centerDir = VecPosition::normalizeAngle( cellAng );
  centerDist = ringRadius;
  gridCenter = agPos;
  relativeCenter = VecPosition::getVecPositionFromPolar( ringRadius, cellAng );
  globalCenter = agPos + relativeCenter;
  cellArea = area;
  cellRadius = sqrt( cellArea/M_PI );
  oFastestTeammate = OBJECT_ILLEGAL;
  iCyclesBallToReach      = 6000;
  iCyclesFastestOpponent  = 6000;
  iCyclesFastestTeammate  = 6000;
  dImportance = 0.0;
  dExpectedOppFactor = 0.0;
  dExpectedTmmFactor = 0.0;

  // determine sector type for this cell
  sectorType = SECTOR_UNKNOWN;
  AngDeg AngDiff = VecPosition::normalizeAngle( cellAng - agDir );
  if ( abs( AngDiff ) < 45 )
    sectorType = SECTOR_FRONT;
  else if ( AngDiff > 0 )
    if ( AngDiff < 135 )
       sectorType = SECTOR_RIGHT;
    else
      sectorType = SECTOR_BACK;
  else
    if ( AngDiff > -135 )
       sectorType = SECTOR_LEFT;
    else
      sectorType = SECTOR_BACK;

} // initCell


AngDeg GridCell::getCellDir()
{
  return centerDir;
}


double GridCell::getCellDist()
{
  return centerDist;
}

void GridCell::eliminate()
{
  isActive = false;
}

bool GridCell::isInUse()
{
  return isActive;
}

VecPosition GridCell::getCellCenter()
{
  return globalCenter;
}

VecPosition GridCell::getGridCenter()
{
  return gridCenter;
}

double GridCell::getCellEffRadius()
{
  return cellRadius;
}

double GridCell::getCellArea()
{
  return cellArea;
}

void GridCell:: setTmmFactor ( double numTmm )
{
  dExpectedTmmFactor = numTmm;
}

double GridCell::getTmmFactor()
{
  return dExpectedTmmFactor;
}

void GridCell:: setOppFactor ( double numOpp )
{
  dExpectedOppFactor = numOpp;
}

double GridCell::getOppFactor()
{
  return dExpectedOppFactor;
}

void GridCell:: setImportance ( double value )
{
  dImportance = value;
}

double GridCell:: getImportance ( )
{
  return dImportance;
}


void GridCell:: setCyclesBallToReach    ( int cyc )
{
  iCyclesBallToReach = cyc;
}

int GridCell:: getCyclesBallToReach()
{
  return iCyclesBallToReach;
}

void GridCell:: setCyclesFastestTmm     ( int cyc )
{
  iCyclesFastestTeammate = cyc;
}

int GridCell:: getCyclesFastestTmm()
{
  return iCyclesFastestTeammate;
}


void GridCell:: setCyclesFastestOpp     ( int cyc )
{
  iCyclesFastestOpponent = cyc;
}

int GridCell:: getCyclesFastestOpp()
{
  return iCyclesFastestOpponent;
}


void GridCell:: setFastestTeammate      ( ObjectT tmm )
{
  oFastestTeammate = tmm;
}

ObjectT  GridCell:: getFastestTeammate( )
{
  return oFastestTeammate;
}

GridSectorT GridCell::getSectorType( )
{
  return sectorType;
}


//! This method returns player "presense" factor in this grid cell
//  The angular deviation and geometrical distances between the cell
//  center and the player position are combined in this value.
//    \param 'pos' is the player global position
//    \param 'teammate' is true if it is a teammate
double GridCell:: calcPresenseFactor ( VecPosition  pos, bool isTeammate )
{
  double value;
  double dist = pos.getDistanceTo( globalCenter );

//    value = 0.75 * dist / cellRadius;
    value = 0.75 * dist / 2.0;
    value = 1.0 / ( 1.0 + value * value * value );
                          // at 0.5*cellRadius distance, we get  512/539 = 0.9499
                          // at     cellRadius distance, we get  64/91 = 0.7033
                          // at 1.5*cellRadius distance, we get  512/1241 = 0.4126
                          // at 2.0*cellRadius distance, we get  8/35 = 0.2286
                          // at 3.0*cellRadius distance, we get  64/793 = 0.0807
                          // at 4.0*cellRadius distance, we get  1/28 = 0.0357

    // now consider the angular deviation if this cell is blocked
    double angleFactor = 0.0;
    if ( dist < centerDist )
    {
      AngDeg posDir = ( pos - gridCenter ).getDirection();
      AngDeg angDeviation = abs( VecPosition::normalizeAngle( posDir - centerDir ) );
      double angleThreshold = 3.0;       // magic number
      if ( angDeviation < 3.0 * angleThreshold )
      {
        angleFactor = angDeviation / angleThreshold;
        angleFactor = 1.0 /( 1.0 + angleFactor * angleFactor * angleFactor );
      }
    }

    if ( isTeammate )
      return value - angleFactor / ( 1.0 + dist*dist/100.0 );    // reduce presence in a shadowed cell
    else
      return value + angleFactor;                                // increase presence
}

/*
//! This method returns player "presense" factor in this grid cell
//  The angular deviation and geometrical distances between the cell
//  center and the player position are combined in this value.
//    \param pos is the player global position
  // if the player is located in the cell center, we get 1.00
  // with 0.5*angleThreshold angular deviation and
  // 1.0*cellRadius distance from the cell center, we get 0.8203
double GridCell:: calcPresenseFactor ( VecPosition  pos )
{
  double value;
  double dist = pos.getDistanceTo( globalCenter );

  if ( dist < 2.0 * cellRadius )
  {
    value = 0.5 * dist / cellRadius;
    value = value * value * value;
    value = 1.0 - value;  // at 0.5*cellRadius distance, we get 0.96875
                          // at     cellRadius distance, we get 0.9375
                          // at 1.5*cellRadius distance, we get 0.578125
                          // at 2.0*cellRadius distance, we get 0.0

    // now consider the angular deviation
    AngDeg posDir = ( pos - gridCenter ).getDirection();
    AngDeg angDeviation = abs( VecPosition::normalizeAngle( posDir - centerDir ) );
    double angleThreshold = 10.0;       // magic number
    double angleFactor = 1.0;

    if ( angDeviation < angleThreshold )
    {
      angleFactor = angDeviation / angleThreshold;
      angleFactor = 1.0 - angleFactor * angleFactor * angleFactor;
                          // at 0.5*angleThreshold deviation, we get 0.875
                          // at     angleThreshold deviation, we get 0.0
    }
    else
      angleFactor = 0;

    value = value * angleFactor;
  }
  else
    value = 0.0;

  return value;

}
*/

/**********************  END CLASS GRIDCELL   *************************************/


//! This method intitializes gridcell variables used for for
//  decision making with the ball. Cell attribute values depend on situation.
//    \returns true if any unblocked cells exist
bool Player::initGrid( int *numOfAlternatives )
{
  // initialize regular cell grid
  // we actually are interested in the cell center and its area,
  // without precise knowledge of its shape and borders;
  // however, in these calcualations we treat cells as sectors

  bool output = true;

  int cellCount = 0;
  double maxRadius = 0.0;
  double rMin0 = PLAYER_CIRCLE_RADIUS;
  double rMax = rMin0;                  // cell ring maximal radius
  double rMin = 0.0;                    // cell ring minimal radius

  // iterate rings
  for ( int iRing = 0;  cellCount < MAX_GRIDCELLS;  iRing++ )
  {
    rMax = rMax + (rMax - rMin) * GRIDCELL_FACTOR;
    rMin = rMin0;
    rMin0 = rMax;
    // cell center distance relative to the agent
    double ringRadius = ( rMin + rMax )/2.0;

    //cout << "ringRadius = " << ringRadius << endl;

    if ( maxRadius < ringRadius )
      maxRadius = ringRadius;

    AngDeg dAng = 360.0/MAX_GRIDDIRECTIONS;
    double cellArea = ( rMax*rMax - rMin*rMin ) * M_PI * dAng/180.0;

    // iterate angle
    for ( AngDeg ang = 0.0;
          ang < 360.0 && cellCount < MAX_GRIDCELLS; ang = ang + dAng )
    {
      gridcells[ cellCount ].initCell( WM->getAgentGlobalPosition(),
                                       WM->getAgentGlobalBodyAngle(),
                                       ringRadius, ang, cellArea );
      cellCount++;
    }

  }

  // eliminate cells having no chances
  int elimCellCount = 0;
  if ( eliminateGridCells( maxRadius, &elimCellCount ) )
  {
    // evaluate tactical importance of the remaining cells
    evaluateImportance();

    // evaluate time balance and eliminate poor alternatives
    int elimtime = 0;
    calcGridTimeBalance( &elimtime );
    elimCellCount = elimCellCount + elimtime;

    // create additional cells, if possible
    if ( elimCellCount > 0 )
    {
      int addedCellCount = addMoreImportantCells( elimCellCount );
      elimCellCount = elimCellCount - addedCellCount;
    }
    else
    {
      output = false;
    }
  }
  else
  {
    // all grid cells have been eliminated
    output = false;
  }

  //putGridStatistics();

  *numOfAlternatives = MAX_GRIDCELLS-elimCellCount;

  return output;
}


// for testing purposes only
void Player:: putGridStatistics()
{
  int cellCount = 0;
  int dribbleCount = 0;
  for ( int i = 0; i < MAX_GRIDCELLS; i++ )
  {
    if ( gridcells[ i ].isInUse() )
    {
      cellCount++;
      if ( gridcells[ i ].getFastestTeammate() == WM->getAgentObjectType() )
      {
        dribbleCount++;
      }
    }
  }

  cout << "-- Alternatives remained: " << cellCount
       << "  Of them, to dribble: " << dribbleCount
       << endl;
}



//!  This method re-activates eliminated cells by moving them in the neighbourhood
//   of the most tactically important cells having good time balance.
//   This allows to identify additional attractive alternatives with
//   only slightly different parameters and then investigate them.
//     \param elimcount is the mumber of eliminated cells
//     \return number of cells re-activated
int Player:: addMoreImportantCells( int elimcount )
{
  int activatedCount = 0;

  // NOT IMPLEMENTED YET

  return activatedCount;
}


//! This method eliminates grid cells which make no sense.
//    \returns true if any unblocked cells exist
//    \param maxDist is the maximal distance to grid cells
//    \param  eliminatedCount is the retuned number of eliminated alternatives
bool Player::eliminateGridCells( double maxDist, int *eliminatedCount )
{
  bool output = true;
  maxDist = 1.3 * maxDist;
  int mycount = 0;
  double halfWidth  = PITCH_WIDTH/2.0 - 3.0;
  double halfLength = PITCH_LENGTH/2.0 - 3.0;

  // iterate cells and eliminate those lying outside the pitch
  for ( int i = 0; i < MAX_GRIDCELLS; i++ )
  {
    if ( abs( gridcells[ i ].getCellCenter().getX() ) > halfLength ||
         abs( gridcells[ i ].getCellCenter().getY() ) > halfWidth    )
    {
      gridcells[ i ].eliminate();
      mycount++;
/*
      if ( WM->getCurrentCycle() < 5 )
      cout << " cell " << gridcells[ i ].getCellCenter()
           << " is outside. halfLength=" << halfLength
           << " halfWidth=" << halfWidth << endl;
*/
    }
  }

  //cout << "eliminated lying outside: " << mycount << endl;

  // iterate opponent players and eliminate cells blocked by them
  int         iIndex;
  double      dConfThr      = PS->getPlayerConfThr();
  int         blockedcount = 0;
  ObjectT histype = OBJECT_ILLEGAL;

  for( ObjectT o = WM->iterateObjectStart( iIndex, OBJECT_SET_OPPONENTS, dConfThr );
       o != OBJECT_ILLEGAL;
       o = WM->iterateObjectNext ( iIndex, OBJECT_SET_OPPONENTS, dConfThr ) )
  {
    VecPosition oppPos = WM->getGlobalPosition( o );
    double oppDist = WM->getAgentGlobalPosition().VecPosition::getDistanceTo( oppPos );

    if ( oppDist < maxDist )
    {
      AngDeg oppDir =  (oppPos - WM->getAgentGlobalPosition() ).VecPosition::getDirection();
      // iterate cells and eliminate blocked ones
      for ( int i = 0; i < MAX_GRIDCELLS; i++ )
      {
        if ( gridcells[ i ].isInUse() )
        {
          AngDeg dAng = 5.0;
          AngDeg angDif = VecPosition::normalizeAngle( gridcells[ i ].getCellDir() - oppDir );
          if ( abs ( angDif ) < dAng )
          {
            // this cell direction is less than dAng degrees apart from the opponent direction
            if ( gridcells[ i ].getCellDist() > 0.9*oppDist )
            {
              // this cell distance is greater than, or close to the opponent distance
              // therefore it is blocked
              gridcells[ i ].eliminate();
              blockedcount++;
/*
              cout << "*** cell " << i << " is blocked ***" << endl;
              cout << "cell dist =" << gridcells[ i ].getCellDist()
                   << " cell dir=" << gridcells[ i ].getCellDir()
                   << " radius=" << gridcells[ i ].getCellEffRadius()
                   << " active=" << gridcells[ i ].isInUse()
                   << endl;
              cout
              << " Opp #" << o-72
              << " oppDist=" << oppDist
              << " oppDir=" << oppDir
              << endl;
*/
            }
          }
        }
      }
    }
  }
  WM->iterateObjectDone( iIndex );

  //cout << " eliminated blocked: " << blockedcount << endl;

  int elimcongested = 0;
  if ( *eliminatedCount == MAX_GRIDCELLS )
  {
    // all cells have been eliminated; there is no much sense to continue
    output = false;
  }
  else
  {
    // estimate congestion and eliminate congested cells, if any
    evaluateCongestion( &elimcongested );
    mycount = mycount + blockedcount + elimcongested;

    if ( mycount == MAX_GRIDCELLS )
      output = false;
  }

  *eliminatedCount = mycount;

  //cout << " eliminated congested: " << elimcongested  << endl;

  return output;

} // eliminateGridCells


//! This method evaluates player congestion in the cells of the grid and
//  eliminates high-congested cells
//     \param  elimhere is the retuned number of eliminated alternatives
void Player::evaluateCongestion( int *elimhere )
{
  int eliminatedHere = 0;
  // iterate remaining cells
  for ( int i = 0; i < MAX_GRIDCELLS; i++ )
  {
    if ( gridcells[ i ].isInUse() )
    {
      // calculate congestion in the cell and eliminate it, if appropriate
      if ( eliminateCongestedCell( i ) )
      {
        eliminatedHere++;
      }
    }
  }
  *elimhere = *elimhere + eliminatedHere;
}


//!  This method returns the player "presense" factor for given
//   grid cell and given team.
//     \param idx is the cell index
//     \param set is the object set type
double Player:: evaluatePresense( int idx, ObjectSetT set )
{
  // iterate players and accumulate congestion for this cell
  double presence = 0.0;
  int         iIndex;
  double      dConfThr      = PS->getPlayerConfThr();

  for( ObjectT o = WM->iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = WM->iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    
    double hispresence;
    
    if ( set == OBJECT_SET_OPPONENTS )
    {
      VecPosition oppPos = WM->getGlobalPosition( o );
      hispresence = gridcells[ idx ].calcPresenseFactor( oppPos, false );
    }
    else
    {  
      // for a teammate, move his position in the direction where he is heading to
      double time = gridcells[ idx ].getCellDist()/1.2; // estimated time ball to reach the cell
      if ( time > 4.0 ) time = 4.0;
      VecPosition tmmPos = WM->getGlobalPosition( o ) + WM->getGlobalVelocity( o ) * time;
      hispresence = gridcells[ idx ].calcPresenseFactor( tmmPos, true );
    }    
    presence = presence + hispresence;       // accumulated presence in gridcells[ idx ]
  }
  WM->iterateObjectDone( iIndex );

  return presence;

}



//! This method calculates player congestion for given grid cell and
//  returns true if it has been elinated
bool Player::eliminateCongestedCell( int cellIdx )
{
  bool eliminate = false;

  // estimate opponent congestion for this cell
  double OppPresense = evaluatePresense( cellIdx, OBJECT_SET_OPPONENTS );
  double TmmPresense;
  if ( OppPresense > 0.6 )
  {
    // eliminate cell congested by opponents
    gridcells[ cellIdx ].eliminate();
    eliminate = true;
/*
    cout << "=== cell #" << cellIdx << "   ===  is congested  === "
       << " OpPres=" << OppPresense
       << "  TmPres=" << TmmPresense
       << "  active=" << gridcells[ cellIdx ].isInUse()
       << endl;
*/
  }
  else
  {
    // save congestion factors for this cell
    gridcells[ cellIdx ].setOppFactor( OppPresense );
    // iterate teammates and estimate congestion for this cell
    TmmPresense = evaluatePresense( cellIdx, OBJECT_SET_TEAMMATES );
    gridcells[ cellIdx ].setTmmFactor( TmmPresense );
  }

  return eliminate;

}


//!  This method calculates time parametrs of the cells in
//   the grid and eliminates poor altrnatives
//     \param  elimhere is the retuned number of eliminated alternatives
void Player::calcGridTimeBalance( int *elimhere )
{
  int elimcount = 0;

  for ( int i=0; i < MAX_GRIDCELLS; i++ )
  {
    if ( gridcells[ i ].isInUse() )
    {
      // determine time balance for teamates and opponents
      int iCycOpp;
      ObjectT oOpp = WM->getFastestInSetTo( OBJECT_SET_OPPONENTS,
                                            gridcells[ i ].getCellCenter(),
                                            VecPosition(0.0, 0.0), 1.0, &iCycOpp );
      int iCycTmm;
      ObjectT oTmm = WM->getFastestInSetTo( OBJECT_SET_TEAMMATES,
                                            gridcells[ i ].getCellCenter(),
                                            VecPosition(0.0, 0.0), 1.0, &iCycTmm );

      // check if the fastest teammate is tired
      int cycAfterComplaint = WM->getCurrentCycle() -
                              WM->Teammates[ SoccerTypes::getTeammateNumber( oTmm ) ].getCycleLowStaminaSaid();
      const int cyclesToRecover = 100;
      if ( cycAfterComplaint < cyclesToRecover )
      {
        // assume that tired player just moves slower
        double decay =  2.0 - (double)cycAfterComplaint / (double)cyclesToRecover;
        // ( with 10 cycles after the complaint, mutiply iCycTmm by 1.9 )
        iCycTmm = (int)( (double)iCycTmm * decay );
      }
/*
        cout << "-- cell #" << i
             << " Importance=" << gridcells[ i ].getImportance()
             << " Opp="  << iCycOpp
             << " Tmm="  << iCycTmm
             << " Trecover=" << cycAfterComplaint
             << " #" << oTmm-60
             << endl;
*/

      if ( iCycTmm > iCycOpp - 0 )
      {
        // opponent is faster, or teammate has too little advantage:
        // remove this cell for poor time disbalance
        gridcells[ i ].eliminate();
        elimcount++;
        //cout << " eliminated: opponent is faster" << endl;
      }
      else
      {
        // evaluate the time balance between the ball and the players
        double ballEndSpeed;

        if ( oTmm == WM->getAgentObjectType() )
        {
          // the agent himself is the fastest to this cell:
          // set special ball end speed for dribbling
          ballEndSpeed = PS->getSlowPassEndSpeed();
        }
        else
        {
//          ballEndSpeed = PS->getSlowPassEndSpeed();   // this is the worst-case for remote cells
          ballEndSpeed = PS->getPassEndSpeed();
        }
        // determine time needed for the ball to reach the cell center
        int iCyclesBall = WM->getNrCyclesBallToMove( gridcells[ i ].getGridCenter(),
                                                     gridcells[ i ].getCellCenter(),
                                                     ballEndSpeed                    );
        // add body+ball turning overhead, if applicable
        if ( gridcells[ i ].getSectorType() == SECTOR_LEFT ||
             gridcells[ i ].getSectorType() == SECTOR_RIGHT   )
          iCyclesBall = iCyclesBall + 1;
        else if ( gridcells[ i ].getSectorType() == SECTOR_BACK )
          iCyclesBall = iCyclesBall + 2;

        //cout << " ball="  << iCyclesBall;

        if ( iCyclesBall < iCycTmm ||       // ball arrives before the teammate gets there, OR
             iCyclesBall >= iCycOpp + 0  )   // after the ball arrival, the opponent gets there too soon
        {
          // eliminate this alternative ( ball too fast or too slow )
          gridcells[ i ].eliminate();
          elimcount++;
          //cout << " eliminated: ball too fast or too slow" << endl;
        }
        else
        {
          // all time constraints are satisfied
          gridcells[ i ].setCyclesBallToReach( iCyclesBall );
          gridcells[ i ].setCyclesFastestTmm( iCycTmm );
          gridcells[ i ].setCyclesFastestOpp( iCycOpp );
          gridcells[ i ].setFastestTeammate ( oTmm );
        }
      }
    }
  }

  //cout << " eliminated for time disbalance: " << elimcount << endl;

  *elimhere = elimcount;

} // calcGridTimeBalance


//====== for debugging only  =====
void Player::printGridCells()
{
    cout << endl;

    for ( int i=0; i < MAX_GRIDCELLS; i++ )
    {
      VecPosition mypos = gridcells[ i ].relativeCenter;
      string act = gridcells[ i ].isActive ? "Active" : "Elim";

//      if ( WM->getPlayerNumber() == 1 )
        cout
           << endl
           << "CELL " << i
           << ";  ctrGlb: " << gridcells[ i ].globalCenter
           << " dir = " << gridcells[ i ].centerDir
           << " dist = " << gridcells[ i ].centerDist
           << endl
           //<< " centerRel: " << gridcells[ i ].relativeCenter
           //<< " area = " << gridcells[ i ].cellArea
           << " effr = " << gridcells[ i ].cellRadius
           //<< " secT = " << gridcells[ i ].sectorType
           << " opp = " << gridcells[ i ].dExpectedOppFactor
           << " tmm = " << gridcells[ i ].dExpectedTmmFactor
           << " " << act
           << endl;
    }

//    if ( WM->getPlayerNumber() == 1 )
    cout << endl
         << "TOTAL CELLS = " << MAX_GRIDCELLS
         << ";  rings = "   <<  MAX_GRIDCELLS/MAX_GRIDDIRECTIONS + 1
         << ";  sectors = " << MAX_GRIDDIRECTIONS
         << endl;
}


//! This method evaluates tactical importance factor of the grid cells
void Player::evaluateImportance()
{
  for ( int i = 0; i < MAX_GRIDCELLS; i++ )
  {
    double value = -1000.0;
    if ( gridcells[ i ].isInUse() )
    {
      VecPosition pos = gridcells[ i ].getCellCenter();
      value = calcPosImportance( pos );
    }
    gridcells[ i ].setImportance( value );
  }
}


//! This method evaluates tactical importance factor of given position
//  The factor depends on tactical situation
double Player::calcPosImportance( VecPosition pos )
{
  double value;

  if ( WM->isInOffenseArea( OBJECT_BALL ) )
    value = calcImportanceOffense( pos );
  else if ( WM->isInDefenseArea( OBJECT_BALL )  )
    value = calcImportanceDefense( pos );
  else
    value = calcImportanceMiddle( pos );


  return value * getPassRequestFactor( pos );

}


//! This method evaluates tactical importance of given position
//  in the middle area of the field.
//  Importance is measured by the X-coordiante and the penetration factor.
//  The returned range is between 0.125 and 0.875, if no penetration is expected
  //  for x = -30 we get 0.3571
  //  for x = 30  we get 0.6429
double Player::calcImportanceMiddle( VecPosition pos )
{
  double value = 0.5*( pos.getX() + PITCH_LENGTH )/ PITCH_LENGTH;

  // also consider if the pos lies in the opponent offside (which is good)
  double penetration = pos.getX() - WM->getOffsideX( false );
  double penetrationFactor = 1.0;
  if ( penetration > 0.0 )
    penetrationFactor = 1.0 + 10.0 * penetration/PITCH_LENGTH;
  if ( penetrationFactor > 2.0 )
    penetrationFactor = 2.0;

  return value * penetrationFactor;
}


//! This method evaluates tactical importance of given position
//  in the defense area of the field.
//  Importance is measured by the distance to the own goal and
//  the Y-coordinate.
//  The returned range is between 0.0 and 1.0 (approximately)
  // for x = y = 0  we get 0.0
  // for x = -47.5 and y = 0  ( 10m from the goal )  we get 0.1452
  // for x = -37.5 and y = 0  ( 20m from the goal )  we get 0.5804
  // for x = -49.5 and y = 16 ( 20m from the goal )  we get 0.7737
  // for x = -27.5 and y = 0  ( 30m from the goal )  we get 0.3265
  // for x = 0 and y = 20 ( 20m from the goal )  we get 0.7059
  // for x = 0 and y = 30 ( 30m from the goal )  we get 1.0059
double Player::calcImportanceDefense( VecPosition pos )
{
  double dist = WM->getPosOwnGoal().getDistanceTo( pos );
  double value = 2.0*dist/PITCH_LENGTH;
  value = value*value + 0.6*abs( pos.getY()/PITCH_WIDTH );
  return 4.0*value;
}


//! This method evaluates tactical importance of given position
//  in the offense area of the field.
//  Importance is measured by the distance to the opponent goal and by
//  the Y-coordinate. Goalie location and number of opponents on the way to
//  their goal are two additional contributing factors.
//  The returned range is between 0.0 and 3.0 (approximately, without oppFactor)
  // for x = 47.5 and y = 0  ( 10m from the goal ) and
  //     the goalie is on the bisector, we get 0.855
  //     if the goalie moves halfway to any pole, we get 1.355
  // for x = 49.5 and y = 16 ( 20m from the goal, oppGoalDir = -53.13 degrees) and
  //     the goalie is on the bisector, we get 0.000
  //     if the goalie moves halfway to any pole, we get 0.300
  //     if the goalie is on any pole direction, we get 0.600
  //     if the goalie abandons the goal, we get 1.200
  // for x = 52.5 and y = 20 ( 20m from the goal, oppGoalDir = -90 degrees) and
  //     the goalie is on the bisector (i.e. on the goalline), we get 0.000
  //     even if the goalie moves out of the goal, we still get 0.000
  // for x = 51.5 and y = 19.8998 ( 20m from the goal, oppGoalDir = -84.26 degrees) and
  //     the goalie is on the bisector (i.e. on the goalline), we get 0.3489
  //     if the goalie steps 1m forward, we get 0.4489
  // for x = 50.5 and y = 9.9499 ( 10m from the goal, oppGoalDir = -84.26 degrees) and
  //     the goalie is on the bisector (i.e. on the goalline), we get 0.5037
  //     if the goalie steps 1m forward, we get 0.6037
double Player::calcImportanceOffense( VecPosition pos )
{
  // consider location relative to the opponent goal
  VecPosition oppGoal = WM->getPosOpponentGoal();
  double distToGoal = oppGoal.getDistanceTo( pos );
  double value = 2.0*distToGoal/PITCH_LENGTH;
  value = 1.0 - 4.0 * ( value*value + 0.6*abs( pos.getY()/PITCH_WIDTH ) );
  if ( value < 0.0 )
    value = 0.0;

  AngDeg goalieDir = ( WM->getGlobalPosition( OBJECT_OPPONENT_GOALIE ) - pos ).getDirection();
  AngDeg oppGoalDir = ( oppGoal - pos ).getDirection();
  AngDeg goalieAngDeviation = abs( VecPosition::normalizeAngle( goalieDir - oppGoalDir ) );
  VecPosition leftPole  = oppGoal + VecPosition( 0.0, -6.0 );
  VecPosition rightPole = oppGoal + VecPosition( 0.0,  6.0 );
  AngDeg leftPoleDir  = ( leftPole  - pos ).getDirection();    // no matter which one is right or left
  AngDeg rightPoleDir = ( rightPole - pos ).getDirection();
  AngDeg goalAngHalfSize = abs( VecPosition::normalizeAngle( leftPoleDir - rightPoleDir ) )/2.0;

  // now consider the goalie location
  double goalieFactor;
  if ( WM->getConfidence( OBJECT_OPPONENT_GOALIE ) > 0.9 )
  {
    // if y = 0 and the goalie is on the bisector, we get goalieFactor = 0.0
    // if y = 0 and the goalie is on the pole direction (any), we get 1.0
    // however, for abs(y)>0 this factor decreases because goalAngHalfSize is getting smaller
    goalieFactor = ( goalieAngDeviation/goalAngHalfSize ) * abs( cosDeg( oppGoalDir ) );
    if ( goalieFactor > 2.0 )
      goalieFactor = 2.0;
  }
  else
    goalieFactor = 0.0;
  
  // now count the opponents between this position and the opponent goal:
  int nrOfOpp = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                            2.0 * tanDeg( goalAngHalfSize ), pos, oppGoal );
  double oppFactor;
  if ( nrOfOpp == 0 )
    oppFactor = 3.0;
  else if ( nrOfOpp == 1 )
    oppFactor = 2.0;
  else if ( nrOfOpp == 2 )
    oppFactor = 1.333;
  else
    oppFactor = 1.0;

  return ( value + goalieFactor ) * oppFactor;

}





//! This method determines if a request for pass had been heard, and
//  evaluates given position for the proximity to the interception point.
double Player::getPassRequestFactor( VecPosition pos )
{
  if ( WM->iSayBallMsgCodeHeard == WM->MSG_PASS_ME_BALL )
  {
    double dist = WM->vpSayBallPointHeard.getDistanceTo( pos );
    // range is beween 1.0 and 2.0; for dist = 3.0 we get 1.43
    return 1.0 + 1.0 / ( 1.0 + 0.15*dist*dist );
  }
  else
    return 1.0;
}

//! This method returns true if the agent finds out that he can efficiently move
//  the ball to given position while pasing it to the teammate or dribblin himself.
//  This decision is made on the basis of evaluating differnt alternatives and
//  selecting one which appears to be the best.
//  The method returns false if all the alternatives have proven to be too poor.
//    \param tmate is the object type of the teammate to receive the pass (including agent himself)
//    \param pos is the position to where ball should be passed
//    \param iCycles is the estimated number of cycles to complete the move ball action
bool Player::canImoveBallWell( ObjectT *tmate, VecPosition *pos, int *iCycles )
{
  bool can = true;

  return can;
}



//!  This method uses multiple criteria find the optimal alternative
//   for acting with the ball.
//   It returns false if no good enough alternative ahs been found.
//   The decision is returned as values of the global variables:
//   teammateToPass, and pointToPass. If teammateToPass is the agent
//   himself, the recommended action is dribbling.
//     \param  numOfAlt is the number of the alternatives to consider
bool Player:: optimizeActionWithTheBall( int numOfAlt )
{
  bool found;

  double criteriaArr [ MAX_GRIDCELLS ][ NUM_OF_CRITERIA ];        // the multiple-criteria array

//  cout << "No  idx  time1  time2  Importance Congestion " << endl;

  // fill the criteria array
  for ( int i = 0; i < MAX_GRIDCELLS; i++ )
  {
    for ( int k = 0; k < NUM_OF_CRITERIA; k++ )
      criteriaArr [ i ][ k ] = -3333.33333;           // just fill with anything first...

    criteriaArr [ i ][ 1 ] = gridcells[ i ].getImportance();    // (maximize it)

    int time1 = gridcells[ i ].getCyclesFastestOpp() - gridcells[ i ].getCyclesFastestTmm();
    if ( time1 > 7 )
      time1 = 7;
    criteriaArr [ i ][ 2 ] = time1;                     // time balance 1 (maximize it)

    int time2 =  gridcells[ i ].getCyclesBallToReach() - gridcells[ i ].getCyclesFastestTmm();
    if ( time2 > 4 )
      time2 = 4;
    criteriaArr [ i ][ 3 ] = time2;                     // time balance 2 (maximize it)

    criteriaArr [ i ][ 4 ] = gridcells[ i ].getOppFactor();
    criteriaArr [ i ][ 5 ] = gridcells[ i ].getTmmFactor();

    if ( gridcells[ i ].isInUse() )
    {
      criteriaArr [ i ][ 0 ] = 0.0;                   // clear the alternative elimination flag
/*
      qq++;
      cout << qq << "    " << i
           << "     " << criteriaArr [ i ][ 2 ]
           << "     " << criteriaArr [ i ][ 3 ]
           << "     " << criteriaArr [ i ][ 1 ]
           << "     " << criteriaArr [ i ][ 4 ]
           << endl;
*/
    }
    else
    {
      criteriaArr [ i ][ 0 ] = 1.0;                   // this is an eliminated alternative
    }
  }

  // fill the optimization control parameter array  ( ctlArr[ 0 ] is not used )
  int ctlArr[ NUM_OF_CRITERIA ];
  for ( int k = 0; k < NUM_OF_CRITERIA; k++ ) ctlArr[ k ] = 0;

  // apply criteria in the following order:
  ctlArr[ 1 ] = -4;      // minimize the opponent presense
  ctlArr[ 2 ] =  2;      // maximize time balance 1
  ctlArr[ 3 ] =  5;      // maximize teammate presense
  // the main criterion
  ctlArr[ 4 ] =  1;      // maximize tactical importance of the ball destination

  int numOfCtriteria = 4;

  int idx = WM->getBestAlternativeIdx( criteriaArr, numOfCtriteria,
                                       MAX_GRIDCELLS, ctlArr,  15.0, 50.0 );

  if ( idx >= 0 )
  {
    double currentImportance = calcPosImportance( gridcells[ idx ].getGridCenter() );
    double gain = gridcells[ idx ].getImportance() - currentImportance;
    gain = 2.0 * gain / ( gridcells[ idx ].getImportance() + currentImportance );

    if ( gain < 0.3  &&  WM->getCurrentCycle() > 5 )           
    {
      // the expected tactical gain is too low: use different heuristic methods
      found = false;
/*
        cout << " #" << WM->getPlayerNumber()
             << " -- using heuristics. Cycle=" << WM->getCurrentCycle()
             << endl;
*/
    }
    else
    {
      // ** global ** parameters of the recommended action with the ball:
      teammateToPass = gridcells[ idx ].getFastestTeammate();
      pointToPass    = gridcells[ idx ].getCellCenter();
      found  = true;
/*
      string method = ( teammateToPass == WM->getAgentObjectType() ) ? " DRIBBLE " : " PASS ";
      cout << " #" << WM->getPlayerNumber()
             << " ++ optimization successful. " << method
             << "  Gain = **" << gain
             << "**  Cycle = " << WM->getCurrentCycle()
             << endl;
*/
    }

/*
    string method = ( teammateToPass == WM->getAgentObjectType() ) ? "** DRIBBLE **" : "** PASS **";
    cout << method << " " << WM->getPlayerNumber()
       << " " <<  gridcells[ idx ].getGridCenter()
       << " ===> #" << gridcells[ idx ].getFastestTeammate()-60
       << " " << gridcells[ idx ].getCellCenter()
       << endl
       << "  idx=" << idx
       << " dist = " << gridcells[ idx ].getCellDist()
      // << "  time = " << gridcells[ idx ].getCyclesBallToReach()
       //<< "  dt1 = " << gridcells[ idx ].getCyclesFastestOpp() - gridcells[ idx ].getCyclesFastestTmm()
       //<< "  dt2 = " << gridcells[ idx ].getCyclesBallToReach() - gridcells[ idx ].getCyclesFastestTmm()
       //<< "  opp = " << criteriaArr [ idx ][ 4 ]
       //<< endl
       //<< " v1 = " << calcPosImportance( gridcells[ idx ].getGridCenter() )
       //<< "  v2 = " << gridcells[ idx ].getImportance()
       << "  v2-v1 = ** " << gridcells[ idx ].getImportance() - calcPosImportance( gridcells[ idx ].getGridCenter() )
       << " **"
       //<< " check=" << criteriaArr [ idx ][ 1 ]
       << endl;
*/
  }
  else
  {
    found = false;
    cout << "FAILED to select any optimal action with the ball. Cycle = " << WM->getCurrentCycle() << endl;
  }

  return found;

} // optimizeActionWithTheBall


