/***************************************************************************
                          WorldModelVK.C  -  World model update methods added
                                             by Vadim
                                             All methods based on particle filter
                                             should be removed as unused.
                             -------------------
    begin                : Thu Apr 17 2003
    copyright            : (C) 2003 by Vadim Kyrylov
    email                : vkyrylov@wct
 ***************************************************************************/


#include "WorldModel.h"


/******************************************************************************/
/********************** CLASS WORLDMODEL **************************************/
/******************************************************************************/
 

//! This method returns true if there are no opponent close to the direction
//  of the planned pass
//    \param 'pointToPass'  point considered for passing ball to
//    \param 'sectorWidth'  the width in degrees of the sector taken into consideration
bool WorldModel::isItSafeToPass( VecPosition pointToPass, double sectorWidth )
{
  VecPosition ballpsn = getGlobalPosition( OBJECT_BALL );
  // count the opponents in the sector centered at ballpsn and pointed at pointToPass
  double nrOfOpponents = getNrInSetInCone( OBJECT_SET_OPPONENTS, sectorWidth*0.01, ballpsn, pointToPass );
  return ( nrOfOpponents == 0 );
}


//! This method returns true if there is a teammate near by and it is safe
//  to pass him the ball
//    \param 'teammate' is the teammate suitable to receive the pass
bool WorldModel::isItSafeToPassClose( ObjectT *teammate )
{
  int         iIndex;
  double      dConfThr      = PS->getPlayerConfThr();
  // iterate through teammates and determine the first suitable guy
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_TEAMMATES, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_TEAMMATES, dConfThr ) )
  {
    if ( o == getAgentObjectType() ) continue;         // do not consider yourself
    if ( SoccerTypes::isGoalie( o ) ) continue;        // do not consider the goalie, either
    VecPosition hisPos = getGlobalPosition( o );
    if ( getRelativeDistance( o ) < 8.0 &&
         isItSafeToPass( hisPos, 15.0 )  )         
    {
      // count opponents near the teammate
      if ( getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                               Circle ( hisPos, 3.0 ) ) == 0 )
      {
        // this a suitable teammate
        *teammate = o;
        return true;
      }                            
    }   
  }
  iterateObjectDone( iIndex );
  
  return false;
}


 
//! This method generates audial message about the ball state and
//  the updates of the ball possession, is appropriate.
void WorldModel::prepMsgToUpdatePossesion ( )
{
  bool tryToWarn = true;
  
  if ( not bNeedToSayBall )
  {
    //*** no Say-Ball message prepared by the agent...
    //    finally decide if a Say-Ball message is still needed
    const double maxRadius = 12.0;

    if ( getConfidence( OBJECT_BALL ) > 0.99 )
    {
      // I am sending the message if I have just seen the ball
      if ( getAgentGlobalPosition().getDistanceTo( getBallPos() ) < maxRadius )
      {
        // ... and I am closely located to the ball
        bNeedToSayBall = true;
      }                                                         
    }                                                                              
    else
    {
      // OR, no matter what the confidence is, I am sending if...
//      if ( getPlayerNumber() == getCoordinatorNumber() &&  getCurrentCycle() > 0 )
      if ( getPlayerNumber() == 6  &&  getCurrentCycle() > 0 )
      {
        // ... I am the Coordinator and the Ball possesion state has radically changed
        if ( getBallPossesion() != getTeamPossesion() &&
             getBallPossesion() != POSSESION_NOTHING )
          bNeedToSayBall = true;
          tryToWarn = false;
      }

      if ( not bNeedToSayBall )
      {
        // ... OR if I am just the closest teammate to the ball  
        ObjectT closestTmm = getClosestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL );

        if ( closestTmm == getAgentObjectType() )
        {
          if ( getNrInSetInCircle( OBJECT_SET_TEAMMATES,
                                  Circle( getBallPos(), maxRadius ) ) < 2 )
          {
            // and the only really close one to it
            bNeedToSayBall = true;
          }
        } 
      }
    }

    if ( bNeedToSayBall )
    {
      // now figure out which player is possesing to the ball
      if ( getBallPossesion() == POSSESION_US )
      {
        // a teammate (including myself) is the closest player to the ball
        iSayBallMsgCode = MSG_OUR_BALL;

        if ( tryToWarn )
        {
          // try to override the above message preparation, if necessary
//          tryToWarnAboutOpponent();                                       DISABLED
        }
      }
      else 
      {
        // at least one opponent is closer
        iSayBallMsgCode = MSG_THEIR_BALL;
      }
      // assign other message parameters
      TiredNessT howIamTired = getAgentTiredNess();
      bSayBallLowStamina  = (  howIamTired == TIREDNESS_VERY_BAD );
      // insert info about myself
      vpSayBallPoint = getAgentGlobalPosition();
      iNrCyclesToComplete = 0;
    }
  }

/*  
  if ( getCurrentCycle() > 0 )
  if ( getPlayerNumber() == 2 || getPlayerNumber() == 5 )
    cout << "#" << getPlayerNumber()
         << " " << getPossesionCode ( Ball.getPossesion() )
         << " time = " << getCurrentCycle()
         << endl;
*/

} // prepMsgToUpdatePossesion



//! This method checks if there is a sneaky opponent who is close to 
//  the teammate controlling the ball, and prepares a warning message 
void WorldModel::tryToWarnAboutOpponent()
{
  int iCyclesTmm;
  ObjectT fastestTmm = getFastestInSetTo( OBJECT_SET_PLAYERS, OBJECT_BALL, &iCyclesTmm );
  if ( SoccerTypes::isTeammate( fastestTmm ) && fastestTmm != getAgentObjectType() )
  {
    int iCyclesOpp;
    ObjectT fastestOpp = getFastestInSetTo( OBJECT_SET_OPPONENTS, OBJECT_BALL, &iCyclesOpp );
    if ( getConfidence( fastestOpp ) > 0.97 && iCyclesOpp - iCyclesTmm < 6 )
    {
      // I have reasonably good info and the opponent is close enough
      // now check if the teammate can see the opponent
      VecPosition tmmPos = getGlobalPosition( fastestTmm );
      VecPosition oppPos = getGlobalPosition( fastestOpp );
      // direstion from teammate to the opponent
      AngDeg oppRelativeDir = (oppPos - tmmPos).getDirection();
      AngDeg tmmBodyDir = getGlobalBodyAngle( fastestTmm );
      AngDeg angleDiff = VecPosition::normalizeAngle( oppRelativeDir - tmmBodyDir );

      if (  fabs( angleDiff ) > 85.0 )
      {
        // it is very likely that the teammate cannot see the opponent:
        // override the message code and set the opponent player number
        bNeedToSayBall  = true;
        iSayBallMsgCode = MSG_OPPONENT_CLOSE;
        iPlayerInvolved = SoccerTypes::getIndex( fastestOpp ) + 1;
/* 
        cout << "#" << getPlayerNumber()
             << " MSG_OPPONENT_CLOSE #" << iPlayerInvolved
             << " (" << (int)fastestOpp - OBJECT_TEAMMATE_UNKNOWN << ")"
             << endl;
*/
      }
    }    
  }   
} // tryToWarnAboutOpponent 


 
/*! This method returns number of cycles [0..15] needed the freely
    moving ball to reach the designated point.
    Recommended for using for pass calculations.
    \param pointFrom start point
    \param pointTo   end point
    \param endSpeed  ball velocity in the end point */
int WorldModel::getNrCyclesBallToMove( VecPosition pointFrom,
                           VecPosition pointTo,
                           double endSpeed        )
{
  int nrCycles = -1;
  double speed = endSpeed;
  double dist = pointFrom.getDistanceTo( pointTo );
  while ( dist > 0 && nrCycles < 15 )
  {
    speed = speed/( SS->getBallDecay() );
    dist = dist - speed;
    nrCycles++;
  }
  return nrCycles;
}

  
//!  This method was  moved to here because of significant changes.
bool WorldModel::processPlayerMessage( )
{
  fixBugInWorldModel();            // A PATCH BY VADIM
  processBallStatusMessage();
  
  return true;
}

  
/*! This method processes the communication message from a teammate
    about the ball status and some other non-coordinate info.
    See message encoding conventions in sayBallStatus() [file PlayerVK.C]
    Besides, method WorldModel::storePlayerMessage() assigns the global variables:
      m_strPlayerMsg - message string;
      m_iCycleInMsg - cycle in which message was sent;
      m_timePlayerMsg  - time message was received;
      m_iMessageSender - player ID who sent the message.
    These variables are used in this method.    
    Developed by Vadim                            */ 
bool WorldModel::processBallStatusMessage()
{
  // number of cycles allowed to ignore messages after ball was seen 
  static int  maxDelay;              

  static double maxXY = 50.0;
  static double maxV = 3.0;

  static char strMessage[MAX_MSG];                       // location for message
  char        *strMsg;
  strcpy( strMessage, m_strPlayerMsg );
  strMsg = strMessage;                                   // pointer to work with
  bool updateBallInfo;
  
  
  //** extract the upper two 'bits' from chunk1 **
  int chunk1 = getCharToInt( strMsg[1] );
  if ( chunk1 >= 32 )
  {
    // this is a retransmitted message; do not further retransmit it
    bRetransmitSayBall = false;
    chunk1 = chunk1 - 32;
  }
  else
  {
    // message sent for the first time
    bRetransmitSayBall = true;       // force the retransmission 
  }
  
  if ( chunk1 >= 16 )
  {
    // a message sent by exhausted teammate
    bSayBallLowStaminaHeard = true;
    chunk1 = chunk1 - 16;
  }
  else
  {
    bSayBallLowStaminaHeard = false;
  }  


  //** extract the message type and player ID involved **
  int MsgType = getCharToInt( strMsg[0] );
  int playerID         = 1 + (MsgType-1) % 11;
  iSayBallMsgCodeHeard = (int) ( ( MsgType - playerID ) / 11 ) + 1;


  // now figure out who this message is is all about
  if ( not bRetransmitSayBall                   ||     // retransmitted msg
      iSayBallMsgCodeHeard == MSG_PASSING_TO   ||
      iSayBallMsgCodeHeard == MSG_PASS_ME_BALL ||
      iSayBallMsgCodeHeard == MSG_OPPONENT_CLOSE  )
  {
    iPlayerInvolvedHeard = playerID;
  }
  else
  {
    // otherwise, this is just the messsage sender
    iPlayerInvolvedHeard = m_iMessageSender;
  }

  

  //** extract ball state vector **
  // ball position [0..255]
  int chunk2 = getCharToInt( strMsg[2] );
  int chunk3 = getCharToInt( strMsg[3] );
  int iBallX = convert3in2( chunk1, chunk2, chunk3, 1);
  int iBallY = convert3in2( chunk1, chunk2, chunk3, 2);
  //  ball velocity [0..255]
  chunk1 = getCharToInt( strMsg[4] );
  chunk2 = getCharToInt( strMsg[5] );
  chunk3 = getCharToInt( strMsg[6] );
  int iBallVelX = convert3in2( chunk1, chunk2, chunk3, 1);
  int iBallVelY = convert3in2( chunk1, chunk2, chunk3, 2);
  // translate ball position and velocity back to the initial range
  double dBallX =    convertIntXYtoDblScale( iBallX, 255, maxXY );
  double dBallY =    convertIntXYtoDblScale( iBallY, 255, maxXY );
  double dBallVelX = convertIntXYtoDblScale( iBallVelX, 255, maxV );
  double dBallVelY = convertIntXYtoDblScale( iBallVelY, 255, maxV );

  vpSayBallPosHeard.setVecPosition( dBallX, dBallY );
  vpSayBallVelHeard.setVecPosition( dBallVelX, dBallVelY );

  // extract point of attention (e.g. target area center, or
  // pass interception point, or reported player position)
  double xTA = convertIntXYtoDblScale( getCharToInt( strMsg[7] ), 63, maxXY );
  double yTA = convertIntXYtoDblScale( getCharToInt( strMsg[8] ), 63, maxXY );
  vpSayBallPointHeard.setVecPosition( xTA, yTA );
   

  // extract ball position/velocity confidence and the time parameter
  int iByte9 = getCharToInt( strMsg[9] );
  int conf = iByte9 % 4;

  // here we are taking into the account a one-cycle delay
  // ( i.e. the original 1.00 translates into 0.99 )
  dConfidenceHeard = 0.99 - (double)(conf)/100.0;
  iNrCyclesToCompleteHeard = (int)( iByte9 / 4 );

/*  
  string repeat = bRetransmitSayBall ? "" : " repeatedly";  
  if ( m_iMessageSender == 9 )
//  if ( iSayBallMsgCodeHeard == MSG_OPPONENT_CLOSE )
  cout << "---- hear #"  << getPlayerNumber()
//      << ":  "  << strMsg
      << " from #" << m_iMessageSender
      << " "<< getMsgCodeFromIdx ( iSayBallMsgCodeHeard )
//      << " code = " << iSayBallMsgCodeHeard
      << " plr="<< iPlayerInvolvedHeard
//      << " retr="<< !bRetransmitSayBall
//      << " lowSta="<< bSayBallLowStaminaHeard
//      << endl
//      << " xB="<< vpSayBallPosHeard.getX()
//      << " yB="<< vpSayBallPosHeard.getY()
//      << " vx="<< vpSayBallVelHeard.getX()
//      << " vy="<< vpSayBallVelHeard.getY()
//      << endl
//      << " xTA="<< vpSayBallPointHeard.getX()
//      << " yTA="<< vpSayBallPointHeard.getY()
//      << " cycl="<< iNrCyclesToCompleteHeard
//      << " cnf="<< dConfidenceHeard
//      << " mycnf="<< getConfidence( OBJECT_BALL )
      <<" time = " << getCurrentCycle()
      << repeat
      << endl;
*/
/*
  cout << "hear chunk1="<< chunk1 << " chunk2="<< chunk2
       << " chunk3="<< chunk3  << endl;

  string repeat = bRetransmitSayBall ? "" : " repeatedly";
  if ( iSayBallMsgCodeHeard == MSG_PASSING_TO  &&
       iPlayerInvolvedHeard == getPlayerNumber() )
  cout << "..-> "<< iPlayerInvolvedHeard
       //<< " heard by #" << getPlayerNumber()
       << " cycles = " << iNrCyclesToCompleteHeard
       << " time = " << getCurrentCycle()
       << repeat
       << endl;
*/

    
  //*** to update or not to update? ***
  // update if ball not seen or felt for maxDelay cycles
  maxDelay = conf + 0;     //   ( very critical !!! )
   
  ObjectT objInvolved;
  if ( iSayBallMsgCodeHeard != MSG_OPPONENT_CLOSE )
    objInvolved = OBJECT_BALL;
  else
    objInvolved = SoccerTypes::getOpponentObjectFromIndex( iPlayerInvolvedHeard - 1 );

  bool utilized = false;

  if ( dConfidenceHeard > 0.97 )                  // use only recent info about objInvolved
  {
    int         iDiff = getCurrentCycle() - m_iCycleInMsg; // time difference
    double      dDiff = (double)iDiff/100.0;               // conf difference

    if ( objInvolved == OBJECT_BALL )
    {
      // update the ball info 
      updateBallInfo = ( getTimeChangeInformation( objInvolved ) < getCurrentTime() - maxDelay );

      if ( updateBallInfo )
      {
        if (   bRetransmitSayBall  ||                         // not a retransmission, OR
             ( iPlayerInvolvedHeard != getPlayerNumber() &&   // not about myself
               iSayBallMsgCodeHeard != MSG_PASSING_TO   )   ) // except when I am receiving pass
                 
        {
          if ( getAgentGlobalPosition().getDistanceTo( getBallPos() ) > 5.0)
          {
          // extrapolate heard ball coordinates by 1 cycle
           vpSayBallPosHeard = vpSayBallPosHeard + vpSayBallVelHeard;
           vpSayBallVelHeard = vpSayBallVelHeard * SS->getBallDecay();

          
           processPerfectHearInfoBall( vpSayBallPosHeard,
                                      vpSayBallVelHeard,
                                      dConfidenceHeard - dDiff );
            utilized = true;
          }
        }
      }  
    }
    else
    {
      // update info about the sneaky/close opponent (his coordinates are sent instead the ball's )
      if ( dConfidenceHeard - dDiff > getConfidence( objInvolved )  )
      {
        updateHeardPlayerInfo ( objInvolved, vpSayBallPosHeard + vpSayBallVelHeard,
                                             vpSayBallVelHeard,
                                             dConfidenceHeard - dDiff );
        utilized = true;
      }  
    }                                

    if ( utilized )
    { 
      iNrSayBallUsed++;
/*
      cout << "... heard and utilized by "<< getPlayerNumber()
           <<" time = " << getCurrentCycle()
           << endl;
*/
    }
  }

    
  if ( iSayBallMsgCodeHeard != MSG_PASSING_TO   &&
       iSayBallMsgCodeHeard != MSG_PASS_ME_BALL    )      
  {
    // utilize recent info about the global position of
    // the message sender. This position, however, is not accompanied by
    // the velocity and the standard deviation is about 0.6 meters.
    // Therefore, we keep old velocity 

    double myconf = ( bRetransmitSayBall ) ? 0.99 : 0.98 ;
    ObjectT sender = SoccerTypes::getTeammateObjectFromIndex( m_iMessageSender - 1 );
    if( ( getGlobalPosition( sender ) - vpSayBallPointHeard ).getMagnitude() > 3.5 )
    {
       // I update the sender coordinates if the error is big enough
//       updateHeardPlayerInfo ( sender, vpSayBallPointHeard,
//                               getGlobalVelocity( sender ), myconf );       DOESN'T WORK PROPERLY YET
       iNrTeammatesUpdated ++;
    }
  }            
  
  iNrSayBallHeard++;

  
  // prevent from retransmission of this type of message
  // (to allow more important messages to get through)
  if ( iSayBallMsgCodeHeard == MSG_OUR_BALL ||
       iSayBallMsgCodeHeard == MSG_THEIR_BALL  )
    bRetransmitSayBall = false;


  // process non-coordinate information, if any, here
  processNonCoordinate( );
  
  // prepare the retransmission of the received message
  if ( bRetransmitSayBall )
  {
    iPlayerInvolved     = iPlayerInvolvedHeard;
    iNrCyclesToComplete = iNrCyclesToCompleteHeard;
    iSayBallMsgCode     = iSayBallMsgCodeHeard;
    if ( iSayBallMsgCodeHeard == MSG_PASSING_TO ||
         iSayBallMsgCodeHeard == MSG_PASS_ME_BALL  )
    {
      vpSayBallPoint = vpSayBallPointHeard; 
    }
    else
    {
      // I insert my own position in the retransmitted message;
      // these data could be updated during optimized positioning
      vpSayBallPoint      = getAgentGlobalPosition();
      iNrCyclesToComplete = 0;
    }  
  }
  else
  {
    // clear it for the next message
    iSayBallMsgCode = MSG_NOMESSAGE;
    dConfidenceHeard = 0.0;
  }

  bSayBallLowStamina  = false;    // could be toggled by this agent elsewhere
  
  return true;
  
} // processBallStatusMessage


/*! This method is called when new information about the player is heard.
    It updates the information of the opponent player only when the confidence is higher
    than the information already available in the WorldModel.
    \param plr is the player object type
    \param posGlobal global position of the player
    \param vel global velocity of the player
    \param dConf confidence in the ball information
    \return true when information is updated, false otherwise */
bool WorldModel::updateHeardPlayerInfo ( ObjectT plr, VecPosition posGlobal,
                                         VecPosition velGlobal, double dConf )
{
  if( SoccerTypes::isKnownPlayer( plr ) )
  {
    double oldconf = getConfidence( plr );
    VecPosition oldXY = getGlobalPosition( plr );

    PlayerObject *pob = (PlayerObject*) getObjectPtrFromType( plr );
    Time time = getTimeFromConfidence( dConf );
    pob->setGlobalVelocity       ( velGlobal, time );
    pob->setGlobalPosition       ( posGlobal, time );
    pob->setGlobalPositionLastSee( posGlobal, time );
    pob->setTimeLastSeen         (            time );
    updateObjectRelativeFromGlobal( plr ); 
/*
    if ( iSayBallMsgCodeHeard == MSG_OPPONENT_CLOSE )
      cout << "#" << getPlayerNumber()
           << " heard about #" << SoccerTypes::getIndex( plr ) + 1
           << " dR = " <<  (oldXY - posGlobal).getMagnitude()
           << " "<< getMsgCodeFromIdx ( iSayBallMsgCodeHeard )
           << " oldconf = " << oldconf
           //<< " newconf = " << getConfidence( plr )
           << " cycle = " << getCurrentCycle()
           << endl;
*/
     return true;
  }
  return false;
}
  


//!  This method is used for debugging only
string WorldModel::getPossesionCode ( PossesionT p )
{
  if ( p == POSSESION_US )
  {
    return "POSSESION_US";
  }
  else if ( p == POSSESION_THEM )
  {
    return "POSSESION_THEM";
  }
  else
  {
    return "POSSESION_NOTHING";
  }
}


//!  This method is used for debugging only
string WorldModel::getMsgCodeFromIdx ( int idx )
{
  switch ( idx )
  {
    case -2:
      return "MSG_TAREA_THEM";
      break;
    case -1:
      return "MSG_TAREA_US";
      break;
    case 0:
      return "MSG_NOMESSAGE";
      break;
    case 1:
      return "MSG_PASSING_TO";
      break;
    case 2:
      return "MSG_I_DRIBBLE";
      break;
    case 3:
      return "MSG_OPPONENT_CLOSE";
    case 4:
      return "MSG_PASS_ME_BALL";
      break;
    case 5:
      return "MSG_OUR_BALL";
      break;
    case 6:
      return "MSG_THEIR_BALL";
      break;
    default:
      return "E_R_R_O_R ";
      break;
  }
}



/*! this method updates global variables used in multi-cycle
    actions         */
void WorldModel::processNonCoordinate ( )
{
  bool updatePos = false;

  multiCycleAction = ACT_ILLEGAL; 

  if ( iSayBallMsgCodeHeard == MSG_PASSING_TO )
  {
    iNrCyclesToCompleteInUse = iNrCyclesToCompleteHeard - 1;
    if ( !bRetransmitSayBall ) iNrCyclesToCompleteInUse--;

    if ( iPlayerInvolvedHeard == getPlayerNumber() )
    {
      // this message is for me!
      multiCycleAction = ACT_INTERCEPT_PASS;             // my action for the next several cycles
      // force multi-cycle interception;
//      multiCycleActionCount = 0;                     counter NOT IMPLENETED YET
    }
    else
    {
      // ball passed to a teammate
      multiCycleAction = ACT_GOTO_STRATEGIC_POSITION;     // my action for the next several cycles
    }
  }
  else if ( iSayBallMsgCodeHeard == MSG_PASS_ME_BALL )
  {
    iNrCyclesToCompleteInUse = iNrCyclesToCompleteHeard - 1;
    if ( !bRetransmitSayBall ) iNrCyclesToCompleteInUse--;
    updatePos = false;
    // now figure out if I can execute this request...
    if ( isBallKickable() )
    {
/*
      cout << "#" << getPlayerNumber()
           << " can kick the ball to pass to #" << iPlayerInvolvedHeard
           << " time = " << getCurrentCycle()
           << endl;
*/
      if ( isItSafeToPass( vpSayBallPointHeard, 15.0 ) )                         
      {
        multiCycleAction = ACT_PASS_ON_REQUEST;          
//        cout << " ... and it is safe to pass !!! "<< endl;
      }
    }
  }

  else if ( iSayBallMsgCodeHeard == MSG_OPPONENT_CLOSE )
  // player involved is the opponent, and his ID must be kept
  {
    updatePos = true;                                                        
  }
  else if ( iSayBallMsgCodeHeard == MSG_I_DRIBBLE )
  {
    if ( iPlayerInvolvedHeard == getPlayerNumber() )
    { // this message is for me!  // do nothing, just continue
    } 
    else
    {
      multiCycleAction = ACT_GOTO_STRATEGIC_POSITION;
      updatePos = true;
    }
  }
  else if ( iSayBallMsgCodeHeard == MSG_OUR_BALL ||
            iSayBallMsgCodeHeard == MSG_THEIR_BALL )
  {
//    if ( iPlayerInvolvedHeard == getCoordinatorNumber() )
    if ( iPlayerInvolvedHeard == 6 )
      // syncronize ball possesion status across the team
      if ( iSayBallMsgCodeHeard == MSG_OUR_BALL )
        setTeamPossesion ( POSSESION_US );
      else
        setTeamPossesion ( POSSESION_THEM );
  }                          

  if ( updatePos )
  {
     // save the position and time reported by a teammate (this info is not used anywhere yet)
     Teammates[m_iMessageSender].setPosToMoveToGlobal  ( vpSayBallPointHeard );   
     Teammates[m_iMessageSender].setCycleMoveToPosition( getCurrentCycle() +
                                                         iNrCyclesToCompleteHeard );
  }  

  if ( bSayBallLowStaminaHeard )
  {
    // store info about the complaining teammate  NOTE INDEX IS SMALLER THAN PLAYER NUMBER BY ONE
    Teammates[ m_iMessageSender - 1 ].setCycleLowStaminaSaid( getCurrentCycle() );
  }
                 
} // processNonCoordinate  


/*! this method converts double value from [-zMax, zMax] interval
    to non-negative integer lying in 0..maxint interval
    z=0 maps to maxint/2.0                                        */
int WorldModel::convertDblXYtoIntScale( double z, double zMax, int maxint )
{
  z = min( z,  zMax );                    // prevent from overflow, if any
  z = max( z, -zMax );
  double myZ = ( z + zMax )/(2*zMax);     // translate and scale to (0, 1) first
  return (int)( myZ * maxint + 0.5 );     // round to nearest integer
}

/*! this method converts an integer lying in 0..maxint interval
    to double value in [-zMax, zMax] interval                   */
double WorldModel::convertIntXYtoDblScale( int K, int maxint, double zMax )
{
  // translate and scale to (-1, 1) first
  double myZ = 2.0 * ( (double)K - (double)maxint/2.0 ) / (double)maxint;
  return myZ * zMax;                      // scale to (-zMAx, zMax)
}


/*! this method converts char-encoded value to integer */
int WorldModel::getCharToInt( char mychar )
{
  // Character set for Say message
  // this set was tested on Socceer Server v.9.2.0;
  // total 74 valid character have been detected.
  // Their corresponding integer values are:
  // (first 64 values used for numerical encoding) 48..57, 65..90, 32, 95, 97..122,
  // (misc characters) 40..43, 45, 46, 47, 60, 62, 63.
  // First 64 characters are used for encoding numeric values 0..63
  // (they include digits, letters, space, and underscore, which unlikely will change).
  // The rest 10 characters are used for special cases only; they more likely
  // may be changed in the new Soccer Server versions that will folow.
  const char STR_CHAR_SET[75] =
      "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ _abcdefghijklmnopqrstuvwxyz()*+-./<>?";
  int i;
  for ( i=0; i<75; i++ ) if ( mychar == STR_CHAR_SET[i] ) break;
  return i;
} // getCharToInt


/*! this method converts integer to our propriate char-encoded format */
char WorldModel::getIntToChar( int myint )
{
  char mychar;

  if ( myint >= 0  && myint <= 9 )         // 0..9 to '0'..'9'
    mychar = (char)(myint + '0');
  else if ( myint >= 10  && myint <= 35 )  // 10..35 to 'A'..'Z'
    mychar = (char)(myint-10 + 'A');
  else if ( myint == 36 )                  // 36 to ' '
    mychar = ' ';
  else if ( myint == 37 )                  // 37 to '_'
    mychar = '_';
  else if ( myint >= 38 && myint <= 63 )   // 38..63 to 'a'..'z'
    mychar = (char)(myint-38 + 'a');
  else if ( myint == 64 )                  // 64 to '('
    mychar = '(';
  else if ( myint == 65 )                  // 65 to ')'
    mychar = ')';
  else if ( myint == 66 )                  // 66 to '*'
    mychar = '*';
  else mychar = '?';                       // error

  return mychar;
} // getIntToChar


/*! This method converts three integers, (z1, z2, z3) having 0..63 range
 into two integers (x,y), lying in 0..255 range.
Parameter nr tells if to return x or y.
This is necessary for unpacking data in the Say message   */
int WorldModel::convert3in2( int z1, int z2, int z3, int nr )
{
  switch ( nr )
  {
    case 1: // extract x
      return z1 * 16 + (int)( z2/4 );
    case 2: // extract y
      return (z2 % 4)*64 + z3;
    default: return -1;     // never must get to here
  }
} // convert3in2


/*! This method converts two numbers, (x,y) having (0..255) range
into three (z1, z2, z3), having (0..63) range.
Parameter nr tells if to return z1, z2, or z3.
This is necessary for packing data into the Say message   */
int WorldModel::convert2in3( int x, int y, int nr )
{
  int xlow, yup;
  switch ( nr )
  {
    case 1:
    // z1 contains 4 upper virtual 'bits' of x
      return (int)( x /16 );
    case 2:
    // z2 stores 4 lower virtual 'bits' of x and 2 upper of y
      xlow = x  % 16;
      yup  = (int)( y / 64 );
      return xlow*4 + yup;
    case 3:
      // z3 stores 6 lower 'bits' of y
      return y  % 64;
    default: return -1;     // never must get to here
  }
} // convert2in3



/*! This method recalculates agent global neck angle using the
    the results of first-cut self-localization.
    Developed by Vadim. See the RoboCup'03 research paper for details.
    Works only for two or more visible flags.
    \param posGlobal will be filled with global position of the agent
    \param velGlobal will be filled with global velocity of the agent
    \param angGlobal will be filled with global neck angle of the agent
    \returns corrected neck angle from which systematic error was eliminated. */
double WorldModel::correctedNeckAngle( VecPosition *posGlobal, AngDeg *angGlobal)
{
  double      theDelta = 0, xF1, yF1, xF2, yF2;
  double      xm1, ym1, xm2, ym2, sig2x1, sig2y1, sig2x2, sig2y2;
  double      nominator=0, denominator=0;
  double      deltaX, deltaY, deltaXik, deltaYik, wXik, wYik;
  double      threshold = 0.50; // must be selected rather carefully
  int         iIndex1, iIndex2;

  // for each flag that was perceived in last see message
  for( ObjectT obj1 = iterateObjectStart( iIndex1, OBJECT_SET_FLAGS, 1.0 );
       obj1 != OBJECT_ILLEGAL;
       obj1 = iterateObjectNext ( iIndex1, OBJECT_SET_FLAGS, 1.0 ) )
  {
    // flag Cartesian coordinates
    xF1 = getGlobalPosition  ( obj1 ).getX();
    yF1 = getGlobalPosition  ( obj1 ).getY();
    // estimated agent position using this flag
    xm1 = posAgentFlag[iIndex1].getX();
    ym1 = posAgentFlag[iIndex1].getY();
    // variances of this estimate
    sig2x1 = posAgentFlagSigma2[iIndex1].getX();
    sig2y1 = posAgentFlagSigma2[iIndex1].getY();

    // use pairs with all other flags to calculate the angular bias
    iIndex2 = iIndex1;
    for( ObjectT obj2 = iterateObjectNext ( iIndex2, OBJECT_SET_FLAGS, 1.0 ) ;
         obj2 != OBJECT_ILLEGAL;
         obj2 = iterateObjectNext ( iIndex2, OBJECT_SET_FLAGS, 1.0 ) )
    {
      xF2 = getGlobalPosition  ( obj2 ).getX();
      yF2 = getGlobalPosition  ( obj2 ).getY();
      xm2 = posAgentFlag[iIndex2].getX();
      ym2 = posAgentFlag[iIndex2].getY();
      sig2x2 = posAgentFlagSigma2[iIndex2].getX();
      sig2y2 = posAgentFlagSigma2[iIndex2].getY();

      deltaY = yF1 - yF2;
      if (abs(deltaY) > threshold) {
        deltaXik = (xm2 - xm1)/deltaY;
        wXik = deltaY*deltaY/(sig2x1 + sig2x2);
      } else {
        deltaXik = 0; // in fact, we thus eliminate the "Xik" estimate
        wXik = 0;
      }

      deltaX = (xF1 - xF2);
      if (abs(deltaX) > threshold) {
        deltaYik = (ym1 - ym2)/deltaX;
        wYik = deltaX*deltaX/(sig2y1 + sig2y2);
      } else {
        deltaYik = 0;
        wYik = 0; // we thus eliminate the "Yik" estimate
      }
      denominator = denominator + wXik + wYik;
      nominator = nominator + wXik*deltaXik + wYik*deltaYik;

    }
  }
  if ( abs(denominator)<1e-5 ) denominator = 1e-5;
  {
    // estimated correction to the neck angle
    theDelta = nominator/denominator;
    *angGlobal = *angGlobal + theDelta;
/*
    theDelta = Rad2Deg(theDelta);
    if ( abs(theDelta) > 0.667 ) cout << "******* theDelta = " << theDelta << endl;
*/
  }
  return *angGlobal;
} // correctedNeckAngle



/*! This method calculates agent global position using the
    available information about all visible flags. It uses
    the maximal likelihood technique.
    Developed by Vadim.
    \param posGlobal will be filled with global position of the agent
    \param angGlobal will be filled with global neck angle of the agent
    \returns number of visible flags. */
int WorldModel::updateAgentPosition( VecPosition *posGlobal, AngDeg *angGlobal )
{
  double      x=0.0, y=0.0, xm, ym, delta0, deltax, deltay;
  double      xF, yF, rF, g, rho, sigx, sigy, sigr, coeff;
  double      sig2x, sig2y;
  double      sig2r, sig2g, sing, cosg, sin2g, cos2g;
  double      sqrt12 = sqrt(12.0);
  // standard deviation of angular measurement errors, whose
  //  distribution is uniform in [-0.5, 0.5] degrees
  double      sigg = Deg2Rad(1.0) / sqrt12;


  int         iIndex1, iVisFlagCount = 0;
  double a11=0.0, a12=0.0, a22=0.0;	  // covariation matrix elements
  double c1=0.0, c2=0.0;	            // right-hand parts of equations

  // for all flags that were perceived in last see message
  for( ObjectT objFlag = iterateObjectStart( iIndex1, OBJECT_SET_FLAGS, 1.0 );
       objFlag != OBJECT_ILLEGAL;
       objFlag = iterateObjectNext ( iIndex1, OBJECT_SET_FLAGS, 1.0 ) )
  {
      // get info about the visible flag
      xF = getGlobalPosition  ( objFlag ).getX();
      yF = getGlobalPosition  ( objFlag ).getY();
      rF = getRelativeDistance( objFlag );
      AngDeg gam = getRelativeAngle ( objFlag ) + *angGlobal;
      g = Deg2Rad( gam );

      // standard deviation of distance measurement errors, whose
      //  distribution is uniform in [-0.5%, 0.5%] of the distance, rF
      sigr = 0.01 * rF / sqrt12;

      sig2r = sigr*sigr;
      sig2g = sigg*sigg;
      cosg = cos( g );
      sing = sin( g );
      cos2g = cosg*cosg;
      sin2g = sing*sing;

      // global Cartesian coordinates of the agent
      // calculated using current flag
      xm = xF - rF * cosg;
      posAgentFlag[iIndex1].setX(xm);
      ym = yF - rF * sing;
      posAgentFlag[iIndex1].setY(ym);

      // covariation matrix parameters
      sig2x = sig2r*cos2g + rF*rF*sig2g*sin2g;
      sigx = sqrt( sig2x );
      posAgentFlagSigma2[iIndex1].setX(sig2x);
      sig2y = sig2r*sin2g + rF*rF*sig2g*cos2g;
      sigy = sqrt( sig2y );
      posAgentFlagSigma2[iIndex1].setY(sig2y);
      rho = ( sig2r - rF*rF*sig2g ) * sing * cosg;
      rho = rho /( sigx*sigy );

      coeff = 1 - rho * rho;
      a11 = a11 + 1/( coeff * sigx * sigx );
      a12 = a12 - rho/( coeff * sigx * sigy );
      a22 = a22 + 1/( coeff * sigy * sigy );
      c1 = c1 + ( xm/(sigx * sigx) - rho * ym/(sigx * sigy) )/coeff;
      c2 = c2 + ( ym/(sigy * sigy) - rho * xm/(sigx * sigy) )/coeff;

      iVisFlagCount++;
  }
  iterateObjectDone( iIndex1 );

  if ( iVisFlagCount > 0 )
  {
  // determinants
    delta0 = a11*a22 - a12*a12;
    if ( abs( delta0 ) < 1e-7 )
    { // this should never happen, but them tell...
      cout << "updateAgentPosition: zero determinant " << endl;
      return 0;
    }
    deltax = c1*a22 - c2*a12;
    deltay = a11*c2 - a12*c1;
    // the maximal-likelihood estimate
    x = deltax/delta0;
    y = deltay/delta0;

    posGlobal->setVecPosition( x, y );

  } else {
    cout << "updateAgentPosition: No flags visible to player "
         << getPlayerNumber()
         << endl;
  }
  return iVisFlagCount;
} // updateAgentPosition




/*=================================================================
          Methods for optimizing player positioning
=================================================================*/

/*! this method slightly modifies agent strategic position
    to allow receiving a pass, if the ball is in our posession, or
    to mark the nearest opponent, if the ball is possessed by them, etc.
    - VadimK  */
VecPosition WorldModel::getAdjustedStratPosition( int iPlayer, VecPosition psn )
{
  int caseID = 0;
  VecPosition   posBall  = getBallPos();
  double balldist = posBall.getDistanceTo( psn );  

  if ( balldist < 30.0 )
  {
    // do the adjustments, if any, only in the ball neighbourhood
    if ( isInOffenseArea(OBJECT_BALL) )
    {
      if ( getTeamPossesion()!= POSSESION_THEM )
      {
        // we are attacking
        if ( iPlayer > 5 )
        {
          // adjust forward and midfielder positions
          if ( getPosOpponentGoal().getDistanceTo( psn ) < 15.0 )
            // improve the strategy position to help pass receiving and goal scoring
            caseID = 1;
          else
            // improve the strategy position to help pass receiving and further passing
            // the ball to a teammate closer to the goal
            caseID = 2;
        }
      }
      else  // our attack appears to decay in their defense zone
        if ( iPlayer > 5 )
          // improve the strategy position to help regaining ball control
          caseID = 4;
    }
    else  // ball is on our side or in the middle
    {
      if ( getTeamPossesion() != POSSESION_US )
      {
        // opponent is possessing the ball, or possession is unclear
        if ( iPlayer > 1 )
          // improve the strategy position to help regaining ball control
          caseID = 4;
      }
      else
        // we are possessing the ball
        if ( iPlayer > 1 )
          // improve the strategy position to help ball passing and launching the attack
          caseID = 3;
    }

    // find the optimal position applying different criteria
    if ( caseID > 0 )
      psn = getBestPosition( iPlayer, psn, caseID );
  }

  return psn;

} // getAdjustedStratPosition




/*! This method determine the strategic position for the specified player. This
    is done using the Formations class. In this class all information
    about the current formation, player number in formation and otheic
    values are stored. The strategic position is based on the position of
    the ball. If the confidence in the position of the ball is lower than the
    threshold defined in PlayerSettings, it is assumed that the ball is at
    position (0,0).
    \param iPlayer role in formation for which strategic position should be
      returnd. With default value is -1 it is assumed that the strategic
      position of the agent itself should be returned.
    \param adjust if true, requires position optimization; otherwise,
      only default non-optimized position is returned. - VadimK
    \return VecPosition strategic position for player 'iPlayer'
VecPosition WorldModel::getStrategicPosition( int iPlayer, bool adjust )
{
  VecPosition pos; //, pos1, pos2;
  double dd = 0;

  // -1 is default -> get player number in formation
  if( iPlayer == -1 )
    iPlayer = formations->getPlayerInFormation();

  // get maximal allowed x coordinate, this is offside x coordinate
  double dMaxX = getOffsideX() - 3.0;
  //double dMaxX = getOffsideX() - 1.0;

  // after standing offside we are not allowed to move for ball
  // with a goal kick of them we are not allowed to move into penalty area
  if( isOffsideUs( ) )
    dMaxX = min( dMaxX, getBallPos().getX() - 3.0 );
  if( isGoalKickThem() )
    dMaxX = min( dMaxX, PENALTY_X - 1.0 );
  if( isBeforeKickOff() )
    dMaxX = min( dMaxX, -2.0 );

  // change the ball position on which strategic position is based depending on
  // the different deadball situation and thus the expected movement of the ball
  VecPosition posBall = getBallPos();
  if( isGoalKickUs() ||
      getTimeSinceLastCatch(  ) < PS->getCyclesCatchWait() + 5  )
    posBall.setX( -PITCH_LENGTH/4 + 10.0 );
  else if( getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    posBall.setVecPosition( 0.0, 0.0 );
  else if( isGoalKickThem() ||
           ( isFreeKickThem() && posBall.getX() > PENALTY_X ) )
    posBall.setX( PENALTY_X - 10.0 );

  VecPosition posTarget = getTargetArea();
  VecPosition posPlayer = getAgentGlobalPosition();
  PossesionT ownBall = getTeamPossesion();
  // get the strategic position
  pos = formations->getStrategicPosition1( iPlayer, posBall, posTarget,
                                           posPlayer, ownBall, dMaxX,
                                           isInOffenseArea(OBJECT_BALL) );
  //pos = formations->getStrategicPosition( iPlayer, posBall, dMaxX );
  if (getTimeSinceLastCatch()< 29)
  {
    pos = VecPosition(pos.getX(), pos.getY()+sign(pos.getY())*10.0);
    if (isGoalKickUs())
      pos = VecPosition(pos.getX()-8.0, pos.getY());
  }
  // special positioning for the attack
  if ( isInOffenseArea(OBJECT_BALL) && getTeamPossesion()!=POSSESION_THEM )
  {
    if ( iPlayer > 9 )
    {
      pos = VecPosition(posBall.getX(), pos.getY()*0.6 );
      //   pos = VecPosition(max(dMaxX, posBall.getX()), pos.getY()*0.7 );
    }
   // else if (iPlayer > 6)
   // {
   //   pos = VecPosition(pos., pos.getY() ); //magic number
   // }
  }
  else if (!isInDefenseArea(OBJECT_BALL) && getTeamPossesion()==POSSESION_US )
  {
  }
  else if (isBallWithinRegion(REGION_5a) || isBallWithinRegion(REGION_5b) ||
           isBallWithinRegion(REGION_6a)   )
  {
    dd = 35.0;
    switch(iPlayer)
    {
      case 5:     // player #6
        pos = VecPosition(25.0 + (posBall.getX()-dd)*0.5, 0.0 );
        break;
      case 6:     //player #7
        pos = VecPosition(25.0+(posBall.getX()-dd)*0.9, 7.5 );
        break;
      case 7:     //player #8
        pos = VecPosition(25.0+(posBall.getX()-dd)*0.9, -7.5 );
        break;
      case 8:     // player #9
        pos = VecPosition(35.0 + (posBall.getX()-dd)*0.5, 0.0 );
        break;
      case 9:     //player #A
        pos = VecPosition(35.0+(posBall.getX()-dd)*0.9, 7.5 );
        break;
      case 10:     //player #B
        pos = VecPosition(35.0+(posBall.getX()-dd)*0.9, -7.5 );
        break;
    }
  }
  // // special positioning for the defense
  else if ( isInDefenseArea(OBJECT_BALL) && getTeamPossesion()!=POSSESION_US)
  {
    dd = -PITCH_LENGTH/2.0 + 11.0;
    switch(iPlayer)
    {
      case 1:     // player #2
        pos = VecPosition(-PITCH_LENGTH/2.0+10.0+(posBall.getX()-dd)/2.0, pos.getY()*0.7 );
        break;
      case 2:     //player #3
        pos = VecPosition(-PITCH_LENGTH/2.0+11.0+(posBall.getX()-dd)/2.0,pos.getY()*0.7 );
        //if(posBall.getX() < - PITCH_LENGTH/2.0+10.0)
        //{
        //  cout<<"#3 "<<pos<<" ball "<<posBall<<endl;
        //}
        break;
      case 3:     // player #4
        pos = VecPosition(-PITCH_LENGTH/2.0+11.0+(posBall.getX()-dd)/2.0, pos.getY()*0.7);
        //if(posBall.getX() < - PITCH_LENGTH/2.0+10.0)
        //{
        //  cout<<"#4 "<<pos<<" ball "<<posBall<<endl;
        //}
        break;
      case 4:     // player #5
        pos = VecPosition(-PITCH_LENGTH/2.0+10.0+(posBall.getX()-dd)/2.0, pos.getY()*0.7 );
        break;
      case 5:     // player #6
        pos = VecPosition(pos.getX() - 2.0, pos.getY() );
        break;
      case 6:     // player #7
        pos = VecPosition(pos.getX() - 1.5, pos.getY() );
        break;
      case 7:     // player #8
        pos = VecPosition(pos.getX() - 1.5, pos.getY() );
        break;
    }
  }

  if ( adjust )
  {
    // we are assuming that the position found is more or less good;
    // However, now try to improve it by slightly adjusting it in the neighbourhood of
    // the found default position

            if ( getPlayerNumber() == 6 )
               cout << "*** getStrategicPosition x=" << pos.getX()
                    << ", y="      << pos.getY()
                    << endl;
  
//    pos = getAdjustedStratPosition ( iPlayer, pos );   

            if ( getPlayerNumber() == 6 )
               cout << "... returned getStrategicPosition x=" << pos.getX()
                    << ", y="      << pos.getY()
                    << endl;

  }
  
  return pos;

} // getStrategicPosition
*/



/*! This method first iterates player strategy position over a square grid
    and then analyze them using multiple critria.
    It returns the position which is regarded the best.
    \param iPlayer is the player ID for whom the position is being sought
    \param psn is the default position (the grid center)
    \param caseID is the situation ID (for different situations, different criteria apply)
      - VadimK  */
VecPosition WorldModel::getBestPosition( int iPlayer, VecPosition psn, int caseID )
{
  const int numOfCtriteria = 12;   // maximal number of criteria; to increase it, change all '13', too
  int    iGridHalfNum  =  2;                    // can only be either 1 or 2
                                                // never exceed this limit !!!
  int    stepNum       =  2*iGridHalfNum + 1;
  int    gridsize      =  stepNum * stepNum;

//  double criteriaArr [gridsize][numOfCtriteria + 1];
  // replaced by fixed-size array - to avoid potential OS incompatibilities
  double criteriaArr [ 49 ][ 13 ];        // the multiple-criteria array (7x7 grid maximum)

  double dCellSize     =  PLAYER_REACH/2.0;     // position increment/decrement
                                //====   partial criteria  ====
  double      dNrOppInside;     // number of opponents between the ball and testpsn
  double      dNrOppClose;      // number of opponents very close to testpsn
  double      dNrOppBlocked;    // number of opponents for whom the ball is blocked by testpsn
  double      dNrTmmInside;     // number of teammates between the ball and testpsn
  double      dNrTmmClose;      // number of teammates very close to testpsn
  double      dNrTmmBlocked;    // number of teammates for whom the ball is blocked by testpsn



  // clear them all; this must be coordinated with the control array in getBestPositionIdx()
  for ( int i = 0; i < gridsize; i++ )
  {
    for ( int k = 1; k <= numOfCtriteria; k++ )
    {
      if ( k == 3 || k == 8 )
        criteriaArr[ i ][ k ] = VERY_BIG_NEGATIVE;     // max is sought for these criteria
      else
        criteriaArr[ i ][ k ] = VERY_BIG_POSITIVE;     // min is sought for the rest
    }
  }

  double halfPitchWidth  =  PITCH_WIDTH /2.0;
  double halfPitchLength =  PITCH_LENGTH/2.0;

  double maxgoaldist = 30.0;  // starting from this distance, we count the scoring opportunity

  double xopt            =  psn.getX(),   // default non-adjusted positon
         yopt            =  psn.getY();

  // time for the agent to reach the default position
  int iCyclesToPsn = predictNrCyclesToPoint( getAgentObjectType(), psn,
                                             PS->getPlayerWhenToTurnAngle() );
  VecPosition ballPsn = predictPosAfterNrCycles( OBJECT_BALL, 0 );
  double distance = ballPsn.getDistanceTo( getAgentGlobalPosition() );

  bool found = false;
/*
  cout << "getBestPosition: iCyclesToPsn = " << iCyclesToPsn
       << " Agent:  x = " << getAgentGlobalPosition().getX()
       << " y = " << getAgentGlobalPosition().getY()
       << " Ball:  x = " << ballPsn.getX()
       << " y = " << ballPsn.getY()
       << "  distance = " << distance
       << endl;
*/
  // consider options only within the player and the ball neighbourhoods
  if ( iCyclesToPsn < 10  &&  distance < halfPitchWidth )       // magic numbers
  {
    // iterate positions over the grid and compute the criteria
    int idx = 0;
    for ( int i = -iGridHalfNum; i <= iGridHalfNum; i++ )
    {
      double ypos = psn.getY() + dCellSize * (double)i;
      for ( int j = -iGridHalfNum; j <= iGridHalfNum; j++ )
      {
        double xpos = psn.getX() + dCellSize * (double)j;
/*
        cout << "fabs(xpos) = " << fabs(xpos)
             << " halfPitchLength = " << halfPitchLength
             << "  fabs(ypos) = " << fabs(ypos)
             << " halfPitchWidth = " << halfPitchWidth
             << endl;
*/
        if ( fabs(xpos) < halfPitchLength  &&  fabs(ypos) < halfPitchWidth )
        {
          // consider positions inside the field only
          found = true;

          criteriaArr [ idx ][ 0 ] = 0.0;     // clear the cell elimination flag

          // consider test positions inside the field only
          VecPosition testpsn = VecPosition( xpos, ypos );
          // time for the agent to reach this position
          iCyclesToPsn = predictNrCyclesToPoint( getAgentObjectType(), testpsn,
                                                 PS->getPlayerWhenToTurnAngle() );
          criteriaArr [ idx ][ 10 ] = iCyclesToPsn;

          //*** predict situation for iCyclesToPsn and evaluate it ***

          ballPsn = predictPosAfterNrCycles( OBJECT_BALL, iCyclesToPsn );

          evaluatePlayersForPos( testpsn, iCyclesToPsn, ballPsn, OBJECT_SET_OPPONENTS,
                                          &dNrOppInside, &dNrOppClose, &dNrOppBlocked );

          evaluatePlayersForPos( testpsn, iCyclesToPsn, ballPsn, OBJECT_SET_TEAMMATES,
                                          &dNrTmmInside, &dNrTmmClose, &dNrTmmBlocked );

          criteriaArr [ idx ][ 1 ] = dNrOppInside;       // store these criteria values
          criteriaArr [ idx ][ 2 ] = dNrOppClose;
          criteriaArr [ idx ][ 3 ] = dNrOppBlocked;
          criteriaArr [ idx ][ 4 ] = dNrTmmInside;
          criteriaArr [ idx ][ 5 ] = dNrTmmClose;
          criteriaArr [ idx ][ 6 ] = dNrTmmBlocked;

          if ( caseID < 3 )  //** evaluate the ball-receiving and goal-scoring opportunities
          {
            // evaluate goal-scoring potential
            VecPosition goalpsn = getPosOpponentGoal();
            double goaldist = testpsn.getDistanceTo( goalpsn );
            if ( goaldist < maxgoaldist )
            {
              // count the opponents between testpsn and goalpsn in the sector
              // centered at testpsn (note that test and ball positions are swapped)

              evaluatePlayersForPos( goalpsn, iCyclesToPsn, testpsn, OBJECT_SET_OPPONENTS,
                                              &dNrOppInside, &dNrOppClose, &dNrOppBlocked );
              // count teammates for passing them the ball
              evaluatePlayersForPos( goalpsn, iCyclesToPsn, testpsn, OBJECT_SET_TEAMMATES,
                                              &dNrTmmInside, &dNrTmmClose, &dNrTmmBlocked );

              criteriaArr [ idx ][ 7 ] = dNrOppInside + dNrOppClose;
              criteriaArr [ idx ][ 8 ] = dNrTmmInside + dNrTmmClose;

              if ( caseID == 1 )
              {
                // near the opponent goal: evaluate a two-step ball-passing opportunity
                // from a free teammate, if any
                VecPosition hisPsn;
                ObjectT otmm = getGoodPassingTeammate ( testpsn, iCyclesToPsn, ballPsn, &hisPsn );

                if ( otmm != OBJECT_ILLEGAL )
                {
                  // evaluate congestion in the direction to this teammate
                  evaluatePlayersForPos( testpsn, iCyclesToPsn, hisPsn, OBJECT_SET_OPPONENTS,
                                              &dNrOppInside, &dNrOppClose, &dNrOppBlocked );
                  evaluatePlayersForPos( testpsn, iCyclesToPsn, hisPsn, OBJECT_SET_TEAMMATES,
                                              &dNrTmmInside, &dNrTmmClose, &dNrTmmBlocked );
                  // the congestion index
                  criteriaArr [ idx ][ 9 ] = dNrOppInside + dNrOppClose + dNrTmmInside + dNrTmmClose;
/*
                  cout << "Good passing teammate is #" << (int)otmm << endl;
                  cout << "Opp: Inside = " << dNrOppInside
                       << " Close = " << dNrOppClose
                       << "  Tmm: Inside = " << dNrTmmInside
                       << " Close = " << dNrTmmClose
                       << endl;
*/
                }
              }
            }
          }
        }
        else
        {
          // raise the cell elimination flag
          criteriaArr [ idx ][ 0 ] = 1.0;
        }
/*
        if ( idx == 0 )
        {
          cout << "-------------------------------------------------------------------"
              << endl
              << "                       opponents   "
              << "      teammates   "
              << "    @ goal    passing"
              << endl;
          cout << "idx   xpos  ypos    "
               << "in   cls   blk    "
               << "in   cls   blk    "
               << "op     tm    cong   "
               << "time"

               << endl;
        }

        for ( int q = 0; q < 13; q++ )
        {
          double mynum;
          if ( q == 0 ) mynum = idx;
          if ( q == 1 ) mynum = xpos;
          if ( q == 2 ) mynum = ypos;
          if ( q >= 3 ) mynum = criteriaArr [ idx ][ q - 2 ];
          if ( mynum >  99.0 ) mynum =  99.0;
          if ( mynum < -99.0 ) mynum = -99.0;
          mynum = ( (int)( 100.0*mynum ) )/100.0;

          if (  q == 0 && idx < 10 ) cout << " ";
          if ( mynum >= 0.0 ) cout << " ";
          cout << mynum  << " ";
          if ( fabs( 10.0*mynum - (int)(mynum*10.0  ) ) < 1.e-10 ) cout << " ";
          if ( fabs( mynum - (int)mynum ) < 1.e-10 ) cout << "  ";
        }
        cout << endl;

        if ( idx == gridsize-1 )
          cout << "..................................................................." << endl;
*/
        idx++;
      }
    }
  }

  if ( found )
  {
    // apply differnt criteria to differnt cases and
    // return the best position one-dimensional index in the criteria array
    int posIdx = getBestPositionIdx( caseID, numOfCtriteria, gridsize, criteriaArr );

    if ( posIdx >=0 )
    {
      // get two-dimensional cell index
      int iy = (int)(posIdx / stepNum) - iGridHalfNum;
      int ix = posIdx - (iy + iGridHalfNum) * stepNum - iGridHalfNum;
      // optimized position
      xopt =  xopt + dCellSize * (double)ix;
      yopt =  yopt + dCellSize * (double)iy;
    }
    // else something goes wrong; do nothing
  }
  else
  {
    // nothing better has been found; use the default position
    xopt =  psn.getX(),
    yopt =  psn.getY();
  }

  return VecPosition( xopt, yopt );

} // getBestPosition





/*! This method counts predicted weighted numbers of players located
    in three different zones relative to given position and the ball.
    Counts are weighted with the purpose of making the functions continuous
    \param testpsn is the position to be evaluated
    \param iCycles is the prediction time in cycles
    \param ballPsn is the ball position predicted for iCycles
    \param dNrInside is the returned number of players expected between the ball and testpsn
    \param dNrClose is the returned number of players expected to be in the neighbourhood of testpsn
    \param dNrBlocked is the returned number of players who are blocked by testpsn from the ball
    -  VadimK */
void WorldModel::evaluatePlayersForPos( VecPosition testpsn,  // also could be the opponent goal
                                        int iCycles,
                                        VecPosition ballPsn,  // also could be the evaluated position
                                        ObjectSetT set,
                                        double *dNrInside,
                                        double *dNrClose,
                                        double *dNrBlocked )
{
  double      maxdist  = 25.0;   // magic number (max distance from ballPsn)
  int         iIndex;
  double      dConfThr      = PS->getPlayerConfThr();
  double      ballDist   = ballPsn.getDistanceTo( testpsn );

  // determine the sector of interest; its center (xc, yc) is offset to the opposite
  // side of ballPsn; if the ball is very close, offset is 5 times the player reach;
  // however, if the ball is at max distance, offset is only 1 times player reach

  double offset = PLAYER_REACH * ( 5.0 - 4.0*ballDist/maxdist );
  double cendist = ballDist + offset;            // distance to the center of the sector from testpsn
  double cosA = ( ballPsn.getX() - testpsn.getX() )/(ballDist + 0.01);
  double xc = testpsn.getX() + cendist * cosA;
  double sinA = ( ballPsn.getY() - testpsn.getY() )/(ballDist + 0.01);
  double yc = testpsn.getY() + cendist * sinA;
  VecPosition center = VecPosition ( xc, yc );

  // direction from the sector center to testpsn
  AngDeg centerDegDir = (center - testpsn).getDirection();
  // this narrow sector is about one player reach wide at cendist distance from the center
  AngDeg halfSectorWidth = (AngDeg)( 0.5*57.3*PLAYER_REACH/cendist );
  AngDeg angMin = VecPosition::normalizeAngle( centerDegDir - halfSectorWidth );
  AngDeg angMax = VecPosition::normalizeAngle( centerDegDir + halfSectorWidth );

  // also consider the wider sector (for the continuity)
  AngDeg angMin2 = VecPosition::normalizeAngle( centerDegDir - 2.0*halfSectorWidth );
  AngDeg angMax2 = VecPosition::normalizeAngle( centerDegDir + 2.0*halfSectorWidth );
/*
  cout << "Test position:"
       << " x = " << testpsn.getX()
       << " y = " << testpsn.getY()
       << "  Ball: x = " << ballPsn.getX()
       << " y = "        << ballPsn.getY()
       << " offset = " << offset
       << endl
       << " Wide sector: bisector = " << centerDegDir
       << " angMin2 = " << angMin2
       << " angMax2 = " << angMax2
       << " Narow: angMin = " << angMin
       << " angMax = " << angMax
       << endl;
*/
  //*** predict situation for iCycles and evaluate it ***
  // for the sector centered in the offset of the ball and covering testpsn,
  // count how many players:
  double inside  = 0.0;         // could be blocking testpsn from the ball,
  double close   = 0.0;         //  are within the player reach (congesting his space), and
  double blocked = 0.0;         //  could be blocked from the ball by the player in testpsn

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    if ( o == getAgentObjectType() ) continue;         // do not consider yourself

    // anticipated position of the player
    VecPosition oppPsn = predictPosAfterNrCycles( o, iCycles );

    // distance from the ball to the player, if he was located on the bisector;
    // in other cases this is a bit smaller distance; helps to keep Zone 2 closer to a square;
    // is negative, if the ball is between the player and testpos
    double  oppDist = oppPsn.getDistanceTo( center ) - offset;
/*
    cout << "Player #" << (int)o
         << " x = " << oppPsn.getX()
         << " y = " << oppPsn.getY()
         << " Dist from center = " << oppDist
         << endl;
*/
    // is the player close enough to the ball AND located on my side from it?
    if ( oppDist < maxdist + PLAYER_REACH  &&  oppDist > 0.0 )
    {
      //*** first analyze angles and compute the angle deviation weight
      double angWeight = 0.0;
      AngDeg  oppDegDir = (center - oppPsn ).getDirection();
//      cout << " direction = " << oppDegDir << endl;
      // is the player direction in the wider sector of interest?
      if ( isAngInInterval( oppDegDir, angMin2, angMax2 ) )
      {
//        cout << " is in the wide sector ";
        // is the player in the narrow sector of interest?
        if ( isAngInInterval( oppDegDir, angMin, angMax ) )
        {
          angWeight = 1.0;
//          cout << " (and in narrow)";
        }
        else
        // if the player is beyond the narrow sector,
        // reduce his contribution linearly: if angDiff=halfSectorWidth then angWeight=0.5
        {
          double angDiff = (double)VecPosition::normalizeAngle( centerDegDir - oppDegDir );
          angDiff = fabs( angDiff ) - (double)halfSectorWidth; // deviation from the regular sector
          angWeight = ( 1.0 - 0.5*angDiff/(double)halfSectorWidth );
        }
//        cout << " angWeight = " << angWeight << endl;

        //*** then analyze distances and assign these weights:
        double blockWeight  = 0.0,
               insideWeight = 0.0,
               closeWeight  = 0.0;
        double halfreach = PLAYER_REACH/2.0;
        // distance between testpsn and player
        double deviation = fabs(oppDist-ballDist);

        if ( oppDist > maxdist )
        {
          // this player can likely be blocked from testpsn
          // (outside, but close to Zone3)
          blockWeight = 1.0 - (oppDist - maxdist)/PLAYER_REACH;

          if ( deviation <= PLAYER_REACH )
          {
            // he also contributes to congestion
          if ( deviation < halfreach)
            // this player is for sure congesting the neighbourhood
            closeWeight = 1.0;
          else
            // his congestion factor is reduced up to 0.5 on the border of Zone 2
            closeWeight = 1.0 - 0.5*(deviation-halfreach)/halfreach;
          // however, since it is close to maxdist, this contribution reduces with oppDist
          closeWeight = closeWeight * blockWeight;
          }
/*
          cout << "... got in Zone 3a: insideWeight = " << insideWeight
               << ", closeWeight = " << closeWeight
               << ", blockWeight = " << blockWeight << endl;
*/
        }
        else if ( oppDist > PLAYER_REACH + ballDist )
        { //** zone 3
          // this player can be blocked from the ball for sure
          blockWeight = 1.0;
          if ( oppDist <= PLAYER_REACH + ballDist + halfreach)
            // besides, he could contribute to the congestion ( 0.5 maximum )
            closeWeight = 0.5*(PLAYER_REACH + ballDist + halfreach - oppDist)/halfreach;
/*
          cout << "... got in Zone 3: insideWeight = " << insideWeight
               << ", closeWeight = " << closeWeight
               << ", blockWeight = " << blockWeight << endl;
*/
        }
        else if ( deviation <= PLAYER_REACH )
        { //** zone 2 (almost a square of 2*PLAYER_REACH size)
          if ( deviation < halfreach)
            // this player is for sure congesting the neighbourhood
            closeWeight = 1.0;
          else
            // his congestion factor is reduced up to 0.5 on the border of Zone 2
            closeWeight = 1.0 - 0.5*(deviation-halfreach)/halfreach;

          if ( oppDist-ballDist > halfreach )
            // besides, this player can be somewehat blocked from the ball
            blockWeight =  1.0 - closeWeight;
          else if ( ballDist-oppDist > halfreach )
          // this player can somewehat block the ball for testpsn
            insideWeight = 1.0 - closeWeight;
/*
          cout << "... got in Zone 2: insideWeight = " << insideWeight
               << ", closeWeight = " << closeWeight
               << ", blockWeight = " << blockWeight << endl;
*/
        }
        else if ( ballDist - oppDist > PLAYER_REACH )
        {
          //** Zone 1: between the ball and testpsn
          if ( oppDist > halfreach)
            // this player can block this position from the ball for sure
            insideWeight = 1.0;
          else
            // near the ball, this player contributes less
            insideWeight = 1.0 - (halfreach-oppDist)/halfreach;

          if ( ballDist - oppDist < PLAYER_REACH + halfreach )
            // besides, this player is contributing to the congestion
            closeWeight = -0.5*(ballDist - oppDist - PLAYER_REACH - halfreach)/halfreach;
/*
          cout << "... got in Zone 1: insideWeight = " << insideWeight
               << ", closeWeight = " << closeWeight
               << ", blockWeight = " << blockWeight << endl;
*/
        }

        // keep them all in the 0..1 range (to ensure robustness)
        angWeight    = angWeight < 0.0 ? 0.0 : angWeight;
        angWeight    = angWeight > 1.0 ? 1.0 : angWeight;
        blockWeight  = blockWeight < 0.0 ? 0.0 : blockWeight;
        blockWeight  = blockWeight > 1.0 ? 1.0 : blockWeight;
        insideWeight = insideWeight < 0.0 ? 0.0 : insideWeight;
        insideWeight = insideWeight > 1.0 ? 1.0 : insideWeight;
        closeWeight  = closeWeight < 0.0 ? 0.0 : closeWeight;
        closeWeight  = closeWeight > 1.0 ? 1.0 : closeWeight;

        // output parameters
        blocked = blocked + blockWeight *angWeight;
        inside  = inside  + insideWeight*angWeight;
        close   = close   + closeWeight *angWeight;

      } // angles
    } // ball distance
  }
  iterateObjectDone( iIndex );

  // output parameters
  *dNrInside  = inside;
  *dNrClose   = close;
  *dNrBlocked = blocked;

  return;

} // evaluatePlayersForPos






/*! This method finds best position using multiple criteria and returns its one-dimensional
    index in the grid.
    The alternatives for decision-making in this method are cell indices.
    Different sets of selection criteria apply in different cases.
      \param caseID is the case ID
      \param numOfCriteria is (the second size of the criteria array - 1)
      \param gridsize is the first size of the criteria array
      \param criteriaArr the criteria array.
    -  VadimK */
int WorldModel::getBestPositionIdx( int caseID,    int numOfCriteria,
                                    int gridsize,  double criteriaArr[][13]  )
{
  int  idx;

  // this array is used to set the selection criteria
  int ctlArr[ 13 ];                           // this means that numOfCriteria <= 12
  for ( int i = 0; i <=numOfCriteria; i++ ) ctlArr[i] = 0;


  if ( caseID == 1 )
  {
    // psn is very close to goal; evaluate ball-receiving and goal-scoring opportunities

    // apply criteria in the following order:
    ctlArr[ 1 ] = -10;     // minimize the number of cycles to reach the position
    ctlArr[ 2 ] = -1;      // minimize the number of opponents between the ball and the position
    ctlArr[ 3 ] = -2;      // minimize the number of opponents close to the position
    ctlArr[ 4 ] = -4;      // minimize the number of teammates between the ball and the position
    ctlArr[ 5 ] = -5;      // minimize the number of teammates close to the position
    ctlArr[ 6 ] = -1;      // minimize the number of opponents between the ball and the position
    // the main criterion
    ctlArr[ 7 ] = -7;      // minimize the number of opponents between the goal and the position

    idx = getBestAlternativeIdx( criteriaArr, numOfCriteria, gridsize, ctlArr,  15.0, 15.0 );

    if ( criteriaArr[ idx ][ 1 ] > 1.0 )
    {
      // even the *best* position is blocked by the opponent:
      // now try to find a fairly good position for receiving a two-step pass
      // apply criteria in the following order:
      ctlArr[ 1 ] = -10;     // minimize the number of cycles to reach the position
      ctlArr[ 2 ] = -7;      // minimize the number of opponents between the goal and the position
      // the main criterion
      ctlArr[ 3 ] = -9;      // minimize congestion between the position and the passing teammate

      int newidx = getBestAlternativeIdx( criteriaArr, numOfCriteria, gridsize, ctlArr,  15.0, 25.0 );
      if ( criteriaArr[ newidx ][ 9 ] < 1.0 )
      {
//        cout << "opening for a passing teammate; congestion = " << criteriaArr[ newidx ][ 9 ] << endl;
        idx = newidx;
      }
    }

  }
  else if ( caseID == 2 )
  {
    // not very close to the goal, but rather close anyway
    // evaluate ball-receiving and ball-passing-to-goal opportunities
    // apply criteria in the following order:
    ctlArr[ 1 ] = -10;    // minimize the number of cycles to reach the position
    ctlArr[ 2 ] = -1;     // minimize the number of opponents between the ball and the position
    ctlArr[ 3 ] = -2;     // minimize the number of opponents close to the position
    ctlArr[ 4 ] = -4;     // minimize the number of teammates between the ball and the position
    ctlArr[ 5 ] = -5;     // minimize the number of teammates close to the position
    ctlArr[ 6 ] = -1;     // minimize the number of opponents between the ball and the position
    ctlArr[ 7 ] = -7;     // minimize the number of opponents between the goal and the position
    // the criterion below makes the final choice
    ctlArr[ 8 ] =  8;      // *maximize* the number of teammates between the goal and the position

    idx = getBestAlternativeIdx( criteriaArr, numOfCriteria, gridsize, ctlArr, 15.0, 15.0 );

  }
  else if ( caseID == 3 ) // evaluate ball-receiving and attack-launching opportunities
  {
    // for positions without good chances for scoring
    // apply criteria in the following order:
    ctlArr[ 1 ] = -10;     // minimize the number of cycles to reach the position
    ctlArr[ 2 ] = -2;      // minimize the number of opponents close to the position
    ctlArr[ 3 ] = -1;      // minimize the number of opponents between the ball and the position
    ctlArr[ 4 ] = -4;      // minimize the number of teammates between the ball and the position
    ctlArr[ 5 ] = -5;      // minimize the number of teammates close to the position
    // the criterion below (applied repeatedly here) makes the final choice
    ctlArr[ 6 ] = -1;      // minimize the number of opponents between the ball and the position

    idx = getBestAlternativeIdx( criteriaArr, numOfCriteria, gridsize, ctlArr, 15.0, 15.0 );

  }
  else if ( caseID == 4 ) // evaluate regaining ball control opportunities
  {
    // apply criteria in the following order:
    ctlArr[ 1 ] = -10;    // minimize the number of cycles to reach the position
    ctlArr[ 2 ] = -1;     // minimize the number of opponents between the ball and the position
    ctlArr[ 3 ] = -2;     // minimize the number of opponents close to the position
    ctlArr[ 4 ] = -4;     // minimize the number of teammates between the ball and the position
    ctlArr[ 5 ] = -1;     // minimize the number of opponents between the ball and the position
    ctlArr[ 6 ] = -5;     // minimize the number of teammates close to the position
    // the criterion below makes the final choice
    ctlArr[ 7 ] =  3;     // *maximize* the number of opponents blocked from the ball

    idx = getBestAlternativeIdx( criteriaArr, numOfCriteria, gridsize, ctlArr, 15.0, 15.0 );
  }
  else // do nothing; how did you get here?
    idx = -1;

  return idx;

} // getBestPositionIdx




/*! *** CAN BE REUSED FOR OPTIMIZING DIFFERENT PLAYER ACTIONS ***
    This method finds the index of the best alternative in a
    two-dimensional array of its multiple evaluation crtieria.
    The particular criteria used are specified in yet another control array.
    (See the encoding conventions below.)
    If there is an error in some input data, this method returns -1;

    The optimization is performed in two phases.

    In Phase 1, N particular criteria are applied in a loop, in the order as they
    follow in the control array. Each time one criterion is used to eliminate
    the poorest alternative. However, it is eliminated only if the criterion value
    differs by more than *beta* per cent from the best value of this criterion. The loop
    terminates when either none of the criteria is applicable, or when (1-alpha) per cent
    of all the alternatives have been eliminated.

    In Phase 2, only the last criterion is aplied to select the best alternative of the
    remaining at least *alpha* per cent.

    Method returns the one-dimensional index of the position in the criteria array.
      \parm criteriaArr is the ctiteria array
      \param numOfCriteria is the ctiteria array dimension-1
      \param numOfAlternatives is the criteria array second dimension
      \param ctlArr is the control array
      \param beta is the indiffernce percentage threshold used in Phase 1 (0.0..100.0)
      \param alpha is the percentage of the alternaives reserved for Phase 2 (0.0..100.0).


    Criteria array encoding conventions:
      size: criteriaArr[ numOfAlternatives, numOfCriteria+1 ]
      criteriaArr[ i, 0 ] stores the elimination flag of i-th alternatives; this flag
         must be initialized by the calling method; this allows to eliminate some
         alternatives before applying this optimization algorithm by setting this flag to 1;
         otherwise, it must be set to 0
      criteriaArr[ i, k ] is the k-th criterion value for i-th alternative

    Control array encoding conventions:
      size: ctlArr[ numOfCriteria+1 ]
      ctlArr[ 0 ] is not used
      ctlArr[ k ] = j   means that abs(j)-th criterion must be applied on k-th step;
        this criterion must be maximized if j>0; otherwise, minimized.

    Parameter *beta* equally affects the influence of all the criteria. This
    dependence is inverse: smaller beta results in greater influence.
    The contribution of the last criterion directly depends on *alpha*.
    To significanly increase the contribution of any criterion, it should be listed
    in the control array more than once.

    -  VadimK */
int WorldModel::getBestAlternativeIdx( double criteriaArr[][13],
                                       const int numOfCriteria,
                                       const int numOfAlternatives,
                                       const int ctlArr[],
                                       const double beta,
                                       const double alpha )
{
  //*** preprocess input data ***

  if ( beta < 0.0 || beta >= 100.0 )
  {
    // cancel optimization
    cout << "getBestAlternativeIdx(): fifth parameter value is outside the range [0..100) " << endl;
    return -1;
//<-------
  }

  int numOfCriteriaUsed = 0;
  for ( int k = 1;  k <= numOfCriteria;  k++ )
    if ( abs( ctlArr[ k ] ) > 0 ) numOfCriteriaUsed++;

  if ( numOfCriteriaUsed == 0 )
  {
    // cancel optimization
    cout << "getBestAlternativeIdx(): no criteria have been specified" << endl;
    return -1;
//<-------
  }

  bool isCriterionUsed [ 13 ];     // criterion usage flag
  for ( int k = 1;  k <= numOfCriteriaUsed;  k++ )
    isCriterionUsed[ k ] = true;

  // restrict total number of the alternatives to be eliminated
  // in Phase 1 so that at least one is left for Phase 2.
  int maxOfEliminiated = (int)( ( 1.0 - alpha/100.0 ) * numOfAlternatives + 0.5 );
  if ( maxOfEliminiated >= numOfAlternatives ) maxOfEliminiated = numOfAlternatives - 1;

  // count the alternatives eliminated before calling this function
  int eliminiatedCount = 0;
  for ( int i = 0;  i < numOfAlternatives;  i++ )
   if ( criteriaArr[i][0] > 0.5 )
     eliminiatedCount++;
/*
  cout << "numOfAlternatives = " << numOfAlternatives
       << "  maxOfEliminiated = " << maxOfEliminiated
       << "  eliminiatedCount = " << eliminiatedCount
       << endl;
*/
  if ( eliminiatedCount == numOfAlternatives )
  {
    // cancel optimization
//    cout << "getBestAlternativeIdx(): no alternatives to chose from" << endl;
    return -1;
//<-------
  }


  //*** PHASE 1 ***

  double threshold [ 13 ];     // idividual criteria thresholds
  bool firstLoop = true;
  bool done = false;

  while ( eliminiatedCount < maxOfEliminiated && !done )
  {
    done = true;
    for ( int k = 1;  k <= numOfCriteriaUsed; k++ )
    {
      if ( isCriterionUsed[ k ] )
      {
        // use this criterion to eliminate poor alternatives
        int criterionID = abs( ctlArr[ k ] );

        // find the extremal values and their locations in the array
        int idxMax, idxMin;
        double myMax, myMin;
        findExtremals( criteriaArr, numOfAlternatives, criterionID,
                                    &idxMax, &myMax,  &idxMin, &myMin );

        // use _minimum_ if you want to maximize this criterion
        bool findMin = ctlArr[ k ] > 0 ? true : false;
        int extrIdx = findMin ? idxMin : idxMax;

        if ( firstLoop )
        {
          // set the absolute threshold for k-th criterion
          double range = myMax - myMin;
          threshold[ k ] = findMin ? myMax - range* beta/100.0    // for too big criterion values
                                   : myMin + range* beta/100.0;   // for too small criterion values
/*
          cout << "--- individual threshold for criterion #" << criterionID
              << " is " << threshold[ k ]
              << endl;
*/
        }

        if ( eliminiatedCount < maxOfEliminiated )
        {
          if ( findMin && myMin < threshold[ k ] ||    // we want to eliminate too small values...
              !findMin && myMax > threshold[ k ]   )   // we want to eliminate too big values...
          {
            // at least one criterion is still applicable
            done = false;
            // eliminate the alternative having the extremal value of k-th criterion
            criteriaArr[ extrIdx ][ 0 ] = 1.0;
            eliminiatedCount++;
/*
            cout << "criterion = " << criterionID
                << "  eliminiated idx = " << extrIdx
                << "  total = " << eliminiatedCount
                << endl;
*/
          }
          else
          {
            // do not apply k-th criterion any longer
            isCriterionUsed[ k ] = false;
//            cout << "ctriterion #" << criterionID << " is no longer in use" << endl;
          }
        }
      }
    } // k loop
    firstLoop = false;
  }
/*
  cout << "PHASE 2: choosing from "
       << numOfAlternatives - eliminiatedCount << " alternatives" << endl;
*/

  //*** PHASE 2 ***

  // now find the alternative having the exremal value of the last criterion
  // in the remaining alpha per cent of the alternatives
  int myk = abs( ctlArr[ numOfCriteriaUsed ] );

  // find the extremal values and their locations in the array
  int idxMax, idxMin;
  double maxValue, minValue;
  findExtremals( criteriaArr, numOfAlternatives, myk,
                 &idxMax, &maxValue,  &idxMin, &minValue );

  // use maximum if you want to maximize and minimum otherwise
  int extrIdx = ctlArr[ numOfCriteriaUsed ] > 0 ? idxMax : idxMin;

//  cout << "Best alternative idx = " << extrIdx << endl;

  return extrIdx;

} // getBestAlternativeIdx




/*! This method scans the criteria array and returns maximal and minimal values
    and their indices for given criterion.
*/
void WorldModel::findExtremals( const double criteriaArr[][13],
                                const int arrsize,
                                const int criterionID,
                                int *idxMax, double *maxValue,  int *idxMin, double *minValue )
{
  *minValue = VERY_BIG_POSITIVE;
  *maxValue = VERY_BIG_NEGATIVE;

  for ( int i = 0;  i < arrsize;  i++ )
  {
    if ( criteriaArr[i][0] < 0.5 )
    {
      // not eliminated yet
      if ( *minValue > criteriaArr[ i ][ criterionID ] )
      {
        *minValue = criteriaArr[ i ][ criterionID ];
        *idxMin = i;
      }
      if ( *maxValue < criteriaArr[ i ][ criterionID ] )
      {
        *maxValue = criteriaArr[ i ][ criterionID ];
        *idxMax = i;
      }
    }
  }
} // findExtremals

//-------------------------- end of optimization  ----------------------------


/*! This method tries to find a teammate who has good ball-passing opportunity
    from his location to the position being evaluated. It is assumed that the teammate
    himself has good  chances to receive a pass (which is not checked in this version, though).
      \param testpsn is the position where the teammate will be presumably passing the ball to
      \param iCycles is the prediction time in cycles
      \param ballPsn is the ball position predicted for iCycles
      \param hisPsn is the returned position of the "good" teammate (if any)
    The method returns the teammate object type
*/
ObjectT WorldModel::getGoodPassingTeammate ( VecPosition testpsn, int iCycles,
                                             VecPosition ballPsn, VecPosition *hisPsn )
{
  const double maxdist = 20.0;
  int         iIndex;
  double      dConfThr      = PS->getPlayerConfThr();

  ObjectT histype = OBJECT_ILLEGAL;

  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_TEAMMATES, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_TEAMMATES, dConfThr ) )
  {
    if ( o == getAgentObjectType() ) continue;         // do not consider yourself

    // the anticipated position of the teammate
    *hisPsn = predictPosAfterNrCycles( o, iCycles );
    double hisBallDist  = ballPsn.getDistanceTo( *hisPsn );

    // a "good" player should not be too far away from the ball
    if ( hisBallDist < maxdist  &&
         hisBallDist > PLAYER_REACH )  // but he should not be the guy possessing the ball, anyway
    {
      // also, he must be rather close to testpsn (i.e. to me)
      double hisDistFromMe  = testpsn.getDistanceTo( *hisPsn );
      if ( hisDistFromMe < maxdist &&
           hisDistFromMe > PLAYER_REACH )   // but still not too close
      {
        // and this teammate must be not very close to the straight line connecting testpsn and the ball
        // (because this method is only used when this direction is blocked by opponents)
        AngDeg tb = ( ballPsn - *hisPsn ).getDirection();   // direction teammate -> ball
        AngDeg tp = ( testpsn - *hisPsn ).getDirection();   // direction teammate -> my position
        AngDeg angle = VecPosition::normalizeAngle( tb - tp );
        if ( fabs( angle ) < 150.0 )
        {
          // check if the ball could indeed be smoothly passed to this teammate
          // (not implemented yet)
          histype = o;
          break;
        }
      }
    }
  }
  iterateObjectDone( iIndex );

  return histype;

} // getGoodPassingTeammate


//  This method was created to fix an unknown bug in World Model
//  (In fact, we had no time to find it).
//  In some cases players have zero coordiantes, while the confidence is
//  set to 1.0
//  This method checks all players and corrects confidence, if any player object
//  having zero coordinates was found.   A PATCH BY VADIM
void WorldModel:: fixBugInWorldModel()
{
  int iIndex;
  double dConfThr = 0.0;
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_PLAYERS, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_PLAYERS, dConfThr ) )
  {
    if (  ( VecPosition( 0.0, 0.0 ) - getGlobalPosition( o ) ).getMagnitude() < EPSILON
             && getConfidence( o ) > 0.0    )
    {
/*
      cout << "#" << getPlayerNumber()
           << " zero location of #" << o - 60
           << ". confidence = " << getConfidence( o )
           << " set to zero "
           << " cycle = " << getCurrentCycle()
           << endl;
*/
      PlayerObject *pob = (PlayerObject*) getObjectPtrFromType( o );
      Time time = getTimeFromConfidence( 0.0 );
      pob->setTimeLastSeen( time );          // this results in setting confidence to 0.0
    }       
  }
  iterateObjectDone( iIndex );
}
    

 //  This method finds best interception point lying on given line between two points.
//  Two different methods using different criteria can be used for finding the optimal point.
//  The point sought must be feasible for intercepting the ball in terms of time balance
//  and opponent presence.
//    \param ballPsn is current position of the ball
//    \param ballCycles is the number of cycles for extrapolating ball position before the calculations
//    \param closestPoint is the closest point on the line
//    \param fartherstPoint is the farthest point on the line
//    \param methodID is the method selector
//    \param iMyCycles is the number of cycles needed for the agent to reach the interception point
//    \success is the returned parameter indicating that the sought point exists.
//  author Vadim K.
VecPosition WorldModel::getIntercepionPointOnTheLine( VecPosition ballPsn,
                                                  int ballCycles,
                                                  VecPosition closestPoint,
                                                  VecPosition fartherstPoint,
                                                  int methodID,
                                                  int *iMyCycles,
                                                  bool *success )
{
  bool pointfound = false;
  VecPosition mypoint;
  int iCycles;

  VecPosition ballpsn;
  if ( ballCycles > 0 )
    ballpsn = predictPosAfterNrCycles( OBJECT_BALL, ballCycles );
  else
    ballpsn = ballPsn;

  if ( methodID == 1 )
  {
    // select the farthest point satisfying ball-interception constraints;
    // (this maximizes penetration if the line is directed is towardsd the opponent side)
    mypoint = getFarthestInterceptionPoint( ballpsn, closestPoint, fartherstPoint,
                                                                   &iCycles, &pointfound );
  }
  else if ( methodID == 2 )
  {
    // select the point satisfying ball-interception constraints best;
    // (this maximizes probability of interception)
    mypoint = getSafestInterceptionPoint( ballpsn, closestPoint, fartherstPoint,
                                                                 &iCycles, &pointfound );

  }
  else
  {
    // do nothing
  }

  *iMyCycles = iCycles;
  *success = pointfound;
  return mypoint;

}


//  This method finds the best interception point lying on given line between two points.
//  The best point should be lying as far as possible from the closest point and must
//  satisfy the ball-intercepting time balance and the opponent presence constraints.
//    \param ballPsn is the position of the ball
//    \param closestPoint is the closest point on the line
//    \param fartherstPoint is the farthest point on the line
//    \param iMyCycles is the number of cycles needed for the agent to reach the interception point
//    \success is the returned parameter indicating that the sought point exists.
//  author Vadim K.
VecPosition WorldModel::getFarthestInterceptionPoint( VecPosition ballpsn,
                                               VecPosition closestPoint,
                                               VecPosition fartherstPoint,
                                               int *iMyCycles,
                                                                bool *success )
{
  VecPosition mypoint;             // we want to determine this point
  bool pointfound = false;
  int iCyclesMe = -1000;
  double alpha = 0.0;

  // loop until a reasonably good point is found
  while ( alpha < 1.01 && !pointfound )
  {
    mypoint = fartherstPoint * ( 1.0 - alpha ) + closestPoint * alpha;
    // count the opponents in a ~15-degree sector centered at ballpsn and directed at mypoint
    int nrOfOpponents15 = getNrInSetInCone( OBJECT_SET_OPPONENTS, 0.075, ballpsn, mypoint );
    if ( nrOfOpponents15 == 0 )
    {
      // it appears to be some clear space for passing the ball...
      // now check if I can reach mypoint before the ball
      int iCyclesBall = getNrCyclesBallToMove( ballpsn, mypoint, PS->getSlowPassEndSpeed() );
      iCyclesMe       = predictNrCyclesToPoint( getAgentObjectType(),
                                                  mypoint,
                                                  PS->getPlayerWhenToTurnAngle() );
      if ( iCyclesMe <= iCyclesBall + 3 )    // a tree-cycle tolerance
      {
      // now check if there are any faster opponents to mypoint
        int iCyclesOpp;
        ObjectT fOpp = getFastestInSetTo( OBJECT_SET_OPPONENTS, mypoint,
                                              VecPosition(0.0, 0.0), 1.0, &iCyclesOpp );
        int mytimediff = iCyclesOpp - iCyclesMe;
        if ( mytimediff > 0 )
        {
          // yes, mypoint is fairly good for receiving pass
          pointfound = true;
        }
      }
    }
    // move the point closer to closestPoint ( alpha = 0.0, 0.25, 0.50, 0.75, 1.0 )
    alpha = alpha + 0.25;
  } // while

  *iMyCycles = iCyclesMe;
  *success = pointfound;
  return mypoint;

}


//  This method finds the best interception point lying on given line between two points
//   by comparing several options. The best point should be the most clear of
//  opponents and must satisfy the ball-intercepting time balance constraints.
//    \param ballPsn is the position of the ball
//    \param closestPoint is the closest point on the line
//    \param fartherstPoint is the farthest point on the line
//    \param iMyCycles is the number of cycles needed for the agent to reach the interception point
//    \success is the returned parameter indicating that the sought point exists.
//  author Vadim K.
VecPosition WorldModel::getSafestInterceptionPoint( VecPosition ballpsn,
                                               VecPosition closestPoint,
                                               VecPosition fartherstPoint,
                                               int *iMyCycles,
                                                                bool *success )
{
  bool pointfound = false;
  double alpha = 0.0;
  int opponentFactor = 1000;     // we minimize this criteria, and
  VecPosition bestpoint;         // we want to determine this point


  //cout <<"#" << getPlayerNumber() << " entered getSafestInterceptionPoint... " << endl;

  // iterate through all possible points on the line segment
  while ( alpha < 1.01 )
  {
    VecPosition mypoint = fartherstPoint * ( 1.0 - alpha ) + closestPoint * alpha;
    // count the opponents in a ~10-degree sector centered at ballpsn and directed at mypoint
    int nrOfOpponents10 = getNrInSetInCone( OBJECT_SET_OPPONENTS, 0.05, ballpsn, mypoint );
    if ( nrOfOpponents10 == 0 )
    {

      //cout << "-- no opponents on the ball path" << ", alpha=" << alpha << endl;

      // it appears to be some clear space for passing the ball...
      // now check if I can reach mypoint before the ball
      int iCyclesBall = getNrCyclesBallToMove( ballpsn, mypoint, PS->getSlowPassEndSpeed() );
      int iCyclesMe   = predictNrCyclesToPoint( getAgentObjectType(),
                                                  mypoint,
                                                  PS->getPlayerWhenToTurnAngle() );
      if ( iCyclesMe <= iCyclesBall + 3 )    // a tree-cycle tolerance
      {
        //cout << "-- I can reach the ball" << endl;

        // now check if there are any faster opponents to mypoint
        int iCyclesOpp;
        ObjectT fOpp = getFastestInSetTo( OBJECT_SET_OPPONENTS, mypoint,
                                              VecPosition(0.0, 0.0), 1.0, &iCyclesOpp );
        int mytimediff = iCyclesOpp - iCyclesMe;
        if ( mytimediff > 0 )
        {
          // yes, mypoint is fairly good for receiving pass;  now count the opponents in
          // a ~20-degree sector centered at ballpsn and directed at mypoint
          int nrOfOpponents20 = getNrInSetInCone( OBJECT_SET_OPPONENTS, 0.10, ballpsn, mypoint );

          //cout << "-- I am faster to the ball " << ", nrOfOpponents20=" << nrOfOpponents20 << endl;

          if ( opponentFactor > nrOfOpponents20 )
          {
            // save this by far best point for future use
            pointfound = true;
            bestpoint = mypoint;
            opponentFactor = nrOfOpponents20;
            *iMyCycles = iCyclesMe;
            //cout << "found point " << mypoint << ", alpha=" <<  alpha << endl;
          }
        }
      }
    }
    // move the point closer to closestPoint ( alpha = 0.0, 0.20, 0.40, 0.60, 0.80, 1.0 )
    alpha = alpha + 0.20;
  } // while

  *success = pointfound;
  return bestpoint;

}


//!  This method checks is a through pass is possible and calculates
//   the interception point
bool  WorldModel::canIGiveThroughPass( ObjectT tMate, VecPosition *pos )
{
  bool yesIcan = true;
  VecPosition hisPos = getGlobalPosition(tMate);

  if ( hisPos.getX() < PITCH_LENGTH*0.5 - 20.0 )
  {
    // teammate is too far away from the oponent goal
    yesIcan = false;
  } else if ( hisPos.getDistanceTo( getBallPos() ) > 30.0 )
  {
    // teammate is too far away from the ball
    yesIcan = false;
  }
  else if ( getBallPos().getX() < PITCH_LENGTH*0.5 - 20.0 )
  {
    // the ball is too far away from the oponent goal
    yesIcan = false;
  }
  // I check if it is myself 
  else if ( tMate == getAgentObjectType() )
  {
    yesIcan = false;
  }
  else 
  {        
    cout << "#" << getPlayerNumber()
         << " Trying **through** pass to tMate=" << SoccerTypes::getTeammateNumber( tMate )
         << ". Cycle=" << getCurrentCycle()
         << endl;

    // count the opponents in the narrow sector centered at hisPos and pointed at goalpsn
    const double wideSectorWidth = 16.0;
    VecPosition goalpsn = getPosOpponentGoal();
    double distToGoal = hisPos.getDistanceTo( goalpsn );
    int nrOfOppNarrow = getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                        wideSectorWidth/(4.0*distToGoal), hisPos, goalpsn );
    // count the opponents in a ~30-degree sector centered at hisPos and pointed at goalpsn
    int nrOfOppWide = getNrInSetInCone( OBJECT_SET_OPPONENTS,
                                        wideSectorWidth/(2.0*distToGoal), hisPos, goalpsn );
    double nrOfOpponents = nrOfOppNarrow + 0.5 * nrOfOppWide;

    if ( nrOfOpponents < 2.6 )
    {
      // it looks like there is a good opportunity to score;
      // calculate the ball-inercepting point
      VecPosition mypos;
      double alpha = 0.2;
      bool found = false;
      while ( alpha < 0.6 && not found )
      {
        mypos = hisPos * (1.0-alpha) + getOpponentGoaliePos()*alpha;
        if ( true )  // isItSafeToPass( mypos, 5.0 ) )                     DISABLED
          found = true;
        
        alpha = alpha + 0.1;   
      }  
      yesIcan = found;
      *pos = mypos;
    }
  }
  return yesIcan;
}




//  This method checks if the ball is standing still and updates its position
//  and the counter
void WorldModel::updatePastBallState()
{
  if ( ( getGlobalPosition( OBJECT_BALL ) - oldBallPosition ).getMagnitude() > 0.5 )
  {
    oldBallPosition = getGlobalPosition( OBJECT_BALL );
    iNrCyclesBallIsStill = 0;
  }
  else
  {
    iNrCyclesBallIsStill ++ ;
  }
}


int WorldModel::getCyclesBallStandingStill()
{
  return iNrCyclesBallIsStill;
}

  

